#!/usr/bin/env python
#
# This is a module in development for making conformationally averaged PBE maps
#
debug = False
import sys, os

print __file__
import os

try:
    file_name = __file__
    if file_name[:2] == "./":
        scriptpath = os.getcwd()
    else:
        scriptpath = os.path.join(os.getcwd(), os.path.split(file_name)[0])
        if scriptpath[-1] == "/":
            scriptpath = scriptpath[:-1]
except:
    scriptpath = os.path.split(sys.argv[0])[0]
    if scriptpath == ".":
        scriptpath = os.getcwd()
#
# Add to import path
#
pdb2pqr_path = os.path.split(scriptpath)[0]
sys.path.append(pdb2pqr_path)

import string
import math
import string
import getopt
import time
from src.pdb import *
from src.utilities import *
from src.structures import *
from src.definitions import *
from src.forcefield import *
from src.routines import *
from src.protein import *
from src.server import *
from StringIO import *
from src.hydrogens import *


class conf_avg:
    def __init__(self, options):
        """Initialize class and decide which kind of job to do"""
        self.options = options  # Store options so we can access them anywhere
        potentials = []
        # If directoryPath is specified then use that, otherwise use pdbfilename
        if options.directoryPath != "":
            listOfFiles = os.listdir(options.directoryPath)
            for currentPDB in listOfFiles:
                currentPDB = os.path.join(
                    options.directoryPath, currentPDB
                )  # Jens added this fix
                pots = self.process_one_pdb(currentPDB)
                potentials.append(pots)
        else:
            # Single file
            potentials.append(
                self.process_one_pdb(os.path.join(os.getcwd(), options.pdbfilename))
            )
        #
        # Average potentials
        #
        avg_pots = self.average_potentials(potentials)
        return

    #
    # ------
    #

    def process_one_pdb(self, pdbfilename):
        """Do everything for one input file"""
        print "Working on: %s" % pdbfilename
        pdbfile = getPDBFile(pdbfilename)

        if self.options.MD:
            #
            # Run an MD simulation for this PDB file and calculate potentials for all the snapshots
            #
            snapshots = self.run_MD(pdbfilename)
        else:
            snapshots = [pdbfilename]
        #
        # Get the potentials for everything
        #
        potentials = []
        for pdbname in snapshots:
            pots = self.get_potentials(pdbfilename)
            potentials.append(pdbname)
        return potentials

    #
    # ------
    #

    def run_MD(self, pdbfilename):
        """Run an MD simulation and return a number of snapshots"""
        files = os.listdir(os.getcwd())
        addfiles = [pdbfilename]
        # for file in files:
        # addfiles.append(os.path.join(os.getcwd(),file))
        import Gromacs_class as Gclass

        G = Gclass.GROMACS(addfiles)
        #
        # Create Gromacs input file
        #
        pdbfile = os.path.split(pdbfilename)[1]
        G.pdb2gmx(pdbfile, forcefield=1, ignore_Hs=True, auto_select_his=True)
        #
        # Set up the simulation box. The argument gives the distance between the box edges
        # and the protein in nm
        #
        G.center(1.5)
        G.solvate()
        #
        # Energy minimise and do the pre-MD
        #
        G.EM(2000, 1000)
        params = {}
        params["ref_t"] = "%d" % options.temperature
        G.PR_MD(250)
        # ----------------------------------
        # 500 ps timestep is 2 fs
        G.MD(self.options.MDtime * 500, params)
        #
        # Get the snapshots
        #
        filenames = G.get_snapshots(self.options.numsnapshots)
        return filenames

    def get_potentials(self, currentPDB):
        """Get the potentials by first running pdb2pqr and then apbs"""
        myProtein, apbs_inputfile = self.run_pdb2pqr(currentPDB)
        potentials = self.run_apbs(myProtein, apbs_inputfile)
        return potentials

    def run_pdb2pqr(self, currentPDB):
        """Run pdb2pqr, prepare input for apbs"""
        pdbfile = getPDBFile(currentPDB)
        pdblist, errlist = readPDB(pdbfile)
        #
        # Instantiate pdb2pqr
        #
        myDefinition = Definition()
        myProtein = Protein(pdblist, myDefinition)

        #
        # Setup everything
        #
        myRoutines = Routines(myProtein, verbose)
        myRoutines.updateResidueTypes()
        myRoutines.updateSSbridges()
        myRoutines.updateBonds()
        myRoutines.setTermini()
        myRoutines.updateInternalBonds()

        myforcefield = Forcefield(ff, myDefinition, None)
        myRoutines.applyNameScheme(myforcefield)

        myRoutines.findMissingHeavy()
        myRoutines.addHydrogens()
        myRoutines.debumpProtein()
        myProtein.reSerialize()
        #
        # Add and optimze hydrogens:
        #
        from src.hydrogens import hydrogenRoutines

        myRoutines.updateInternalBonds()
        myRoutines.calculateDihedralAngles()
        myhydRoutines = hydrogenRoutines(myRoutines)
        #
        # Now optimize hydrogens
        #
        myhydRoutines.setOptimizeableHydrogens()
        myhydRoutines.initializeFullOptimization()
        myhydRoutines.optimizeHydrogens()
        myhydRoutines.cleanup()
        myRoutines.setStates()

        print "Created protein object (after processing myRoutines) -"
        print "\tNumber of residues in protein: %s" % myProtein.numResidues()
        print "\tNumber of atoms in protein   : %s" % myProtein.numAtoms()

        #
        # Assign charges
        #
        for chain in myProtein.getChains():
            for residue in chain.get("residues"):
                for atom in residue.get("atoms"):
                    atomname = atom.get("name")
                    charge, radius = myforcefield.getParams1(residue, atomname)
                    atom.set("radius", radius)
                    atom.set("ffcharge", charge)
                #
                #
                method = ""
                async = 0
                split = 0
                import pdb2pka.inputgen_pKa as IP

                igen = IP.inputGen(currentPDB)
                igen.maps = None
                igen.set_type("background")
                igen.pdie = 8.0
                igen.sdie = 80.0
                all_center, extent = igen.getCenter()
                igen.setfineCenter(all_center)
                print "Center: %5.1fA %5.1fA %5.1fA" % (
                    all_center[0],
                    all_center[1],
                    all_center[2],
                )
                print "Extent: %5.1fA %5.1fA %5.1fA" % (extent[0], extent[1], extent[2])

                apbs_inputfile = igen.printInput()
                return myProtein, apbs_inputfile

    def run_apbs(self, myProtein, apbs_inputfile):
        """runs apbs"""
        import pdb2pka.apbs

        APBS = pdb2pka.apbs.runAPBS()
        potentials = APBS.runAPBS(myProtein, apbs_inputfile)
        APBS.cleanup()
        return potentials

    def average_potentials(self, potentials):
        """This function averages many potential maps"""
        avg_pots = []
        for i in range(0, len(potentials[0])):
            currSum = 0
            for j in range(0, len(potentials)):
                currSum += potentials[j][i]
            currAvg = currSum / len(potentials)
            avg_pots.append(currAvg)

        print avg_pots
        return avg_pots


#
# ----
#

if __name__ == "__main__":
    from optparse import OptionParser

    parser = OptionParser(usage="%prog [options]", version="%prog 1.0")
    parser.add_option(
        "-p",
        "--pdb",
        dest="pdbfilename",
        action="store",
        type="string",
        default="2lzt.pka.pdb",
        help="The PDB file. Default: %default",
    )
    parser.add_option(
        "-d",
        "--dir",
        dest="directoryPath",
        action="store",
        type="string",
        default="",
        help="Direcotry of the PDB files/snapshots. Default: %default",
    )
    #
    # Flags
    #
    parser.add_option(
        "--MD",
        dest="MD",
        action="store_true",
        default=False,
        help="Perform an MD simulation and use snapshots for calculating electrostatic potential",
    )
    parser.add_option(
        "-s",
        "--MDsnapshots",
        dest="numsnapshots",
        action="store",
        type="int",
        default=100,
        help="Number of MD snapshots to use. Default: %default",
    )
    parser.add_option(
        "--MDtime",
        dest="MDtime",
        action="store",
        type="int",
        default=100,
        help="Time in picoseconds that MD should be run for. Default: %default",
    )
    parser.add_option(
        "-t",
        "--temp",
        dest="temperature",
        action="store",
        type="float",
        default=310.15,
        help="Temperature for the MD run. Default: %default",
    )
    #
    # We can think about adding flags for not solvating the structure etc here
    #

    (options, args) = parser.parse_args()

    verbose = True
    ff = "parse"

    I = conf_avg(options)



"""
    CGI Server for PDB2PQR

    This module contains the various functions necessary to run PDB2PQR
    from a web server.

    ----------------------------
   
    PDB2PQR -- An automated pipeline for the setup, execution, and analysis of
    Poisson-Boltzmann electrostatics calculations

    Copyright (c) 2002-2011, Jens Erik Nielsen, University College Dublin; 
    Nathan A. Baker, Battelle Memorial Institute, Developed at the Pacific 
    Northwest National Laboratory, operated by Battelle Memorial Institute, 
    Pacific Northwest Division for the U.S. Department Energy.; 
    Paul Czodrowski & Gerhard Klebe, University of Marburg.

	All rights reserved.

	Redistribution and use in source and binary forms, with or without modification, 
	are permitted provided that the following conditions are met:

		* Redistributions of source code must retain the above copyright notice, 
		  this list of conditions and the following disclaimer.
		* Redistributions in binary form must reproduce the above copyright notice, 
		  this list of conditions and the following disclaimer in the documentation 
		  and/or other materials provided with the distribution.
        * Neither the names of University College Dublin, Battelle Memorial Institute,
          Pacific Northwest National Laboratory, US Department of Energy, or University
          of Marburg nor the names of its contributors may be used to endorse or promote
          products derived from this software without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
	ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
	IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
	INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
	BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
	DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
	LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE 
	OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
	OF THE POSSIBILITY OF SUCH DAMAGE.

    ----------------------------
"""

__date__ = "4 August 2008"
__author__ = "Todd Dolinsky, Samir Unni, Yong Huang"

import string
import os
import sys
import time
from aconf import *

# GLOBAL SERVER VARIABLES
# Currently this file is unused. TODO: Nuke it?

""" The absolute path to root HTML directory """
SRCPATH = SRCPATH

""" The relative path to results directory from script directory.
    The web server (i.e. Apache) MUST be able to write to this directory. """
TMPDIR = "tmp/"

""" The maximum size of temp directory (in MB) before it is cleaned """
LIMIT = 500.0

""" The path to the web site *directory* """
WEBSITE = WEBSITE

""" The name of the main server page """
WEBNAME = "index.html"

""" The stylesheet to use """
STYLESHEET = WEBSITE + "pdb2pqr.css"

""" The refresh time (in seconds) for the progress page """
REFRESHTIME = 20

""" The absolute path to the loadavg file - set to "" or None if
    not to be included """
LOADPATH = "/proc/loadavg"

""" The path to the pdb2pqr log - set to "" or None if not to be
    included.  The web server (i.e. Apache) MUST be able to write to
    this directory. """
LOGPATH = "%s/%s/usage.txt" % (INSTALLDIR, TMPDIR)


def setID(time):
    """
    Given a floating point time.time(), generate an ID.
    Use the tenths of a second to differentiate.

    Parameters
        time:  The current time.time() (float)
    Returns
        id  :  The file id (string)
    """
    strID = "%s" % time
    period = string.find(strID, ".")
    id = "%s%s" % (strID[:period], strID[(period + 1) : (period + 2)])
    return id


# def logRun(options, nettime, size, ff, ip):
#    """
#        Log the CGI run for data analysis.  Log file format is as follows:
#
#        DATE  FF  SIZE  OPTIONS   TIME
#
#        Parameters
#            options: The options used for this run (dict)
#            nettime: The total time taken for the run (float)
#            size:    The final number of non-HETATM atoms in the PDB file (int)
#            ff:      The name of the ff used
#            ip:      The ip address of the user
#    """
#    if LOGPATH == "" or LOGPATH == None: return
#    date = time.asctime(time.localtime())
#    debump  = 0
#    opt  = 0
#    propka = 0
#    file = open(LOGPATH,"a")
#
#    if "ffout" in options: ffout = options["ffout"]
#    else: ffout="internal"
#
#    text = "%s\t%s\t%s\t%s\t" % (date, ip, ff, ffout)
#
#    opts = ""
#    if "debump" in options:  opts += "debump,"
#    if "opt" in options:  opts += "optimize,"
#    if "ph" in options: opts += "propka,"
#    if "ligand" in options: opts += "ligand,"
#    if "apbs" in options: opts += "apbs,"
#    if "chain" in options: opts += "chain,"
#
#    if opts == "": opts = "none,"
#
#    text += "%s\t%s\t%.2f\n" % (opts[:-1], size, nettime)
#    file.write(text)
#    file.close()


def cleanTmpdir():
    """
    Clean up the temp directory for CGI.  If the size of the directory
    is greater than LIMIT, delete the older half of the files.  Since
    the files are stored by system time of creation, this is an
    easier task.
    """
    newdir = []
    size = 0.0
    count = 0
    path = INSTALLDIR + TMPDIR

    dir = os.listdir(path)
    for filename in dir:
        size = size + os.path.getsize("%s%s" % (path, filename))
        period = string.find(filename, ".")
        id = filename[:period]
        if id not in newdir:
            newdir.append(id)
            count += 1

    newdir.sort()
    size = size / (1024.0 * 1024.0)

    newcount = 0
    if size >= LIMIT:
        for filename in newdir:
            if newcount > count / 2.0:
                break
            try:
                os.remove("%s%s.pqr" % (path, filename))
            except OSError:
                pass
            try:
                os.remove("%s%s.in" % (path, filename))
            except OSError:
                pass
            try:
                os.remove("%s%s.html" % (path, filename))
            except OSError:
                pass
            newcount += 1


# def getQuote(path):
#    """
#        Get a quote to display for the refresh page.
#        Uses fortune to generate a quote.
#
#        Parameters:
#            path:   The path to the fortune script (str)
#        Returns:
#            quote:   The quote to display (str)
#    """
#    fortune = os.popen(path)
#    quote = fortune.read()
#    quote = string.replace(quote, "\n", "<BR>")
#    quote = string.replace(quote, "\t", "&nbsp;"*5)
#    quote = "%s<P>" % quote
#    return quote

# def printProgress(name, refreshname, reftime, starttime):
#    """
#        Print the progress of the server
#
#        Parameters
#            name:        The ID of the HTML page to write to (string)
#            refreshname: The name of the HTML page to refresh to (string)
#            reftime:     The length of time to set the refresh wait to (int)
#            starttime:   The time as returned by time.time() that the run started (float)
#    """
#    elapsedtime = time.time() - starttime + REFRESHTIME/2.0 # Add in time offset
#    filename = "%s%s%s-tmp.html" % (SRCPATH, TMPDIR, name)
#    file = open(filename,"w")
#    file.write("<HTML>\n")
#    file.write("<HEAD>\n")
#    file.write("<TITLE>PDB2PQR Progress</TITLE>\n")
#    file.write("<link rel=\"stylesheet\" href=\"%s\" type=\"text/css\">\n" % STYLESHEET)
#    file.write("<meta http-equiv=\"Refresh\" content=\"%s; url=%s\">\n" % \
#               (reftime, refreshname))
#    file.write("</HEAD>\n")
#    file.write("<BODY>\n")
#    file.write("<H2>PDB2PQR Progress</H2><P>\n")
#    file.write("The PDB2PQR server is generating your results - this page will automatically \n")
#    file.write("refresh every %s seconds.<P>\n" % REFRESHTIME)
#    file.write("Thank you for your patience!<P>\n")
#    file.write("Server Progress:<P>\n")
#    file.write("<blockquote>\n")
#    file.write("<font size=2>Elapsed Time:</font> <code>%.2f seconds</code><BR>\n" % elapsedtime)
#    file.write("</blockquote>\n")
#    file.write("Server Information:<P>\n")
#    file.write("<blockquote>\n")
#    loads = getLoads()
#    if loads != None:
#        file.write("<font size=2>Server load:</font> <code>%s (1min)  %s (5min)  %s (15min)</code><BR>\n" % (loads[0], loads[1], loads[2]))
#
#    file.write("<font size=2>Server time:</font> <code>%s</code><BR>\n" % (time.asctime(time.localtime())))
#    file.write("</blockquote>\n")
#    file.write("</BODY></HTML>")
#    file.close()

# def printAcceptance(name):
#    """
#        Print the first message to stdout (web browser) - set the
#        refresh to the <id>-tmp.html file.
#
#        Parameters
#            name:    The ID of the HTML page to redirect to (string)
#    """
#    waittime = int(REFRESHTIME/2.0)
#    print "Content-type: text/html\n"
#    print "<HTML>"
#    print "<HEAD>"
#    print "<TITLE>PDB2PQR Progress</TITLE>"
#    print "<link rel=\"stylesheet\" href=\"%s\" type=\"text/css\">" % STYLESHEET
#    print "<meta http-equiv=\"Refresh\" content=\"%s; url=%s%s%s-tmp.html\">" % \
#          (waittime, WEBSITE, TMPDIR, name)
#    print "</HEAD>"
#    print "<BODY>"
#    print "<H2>PDB2PQR Progress</H2><P>"
#    print "The PDB2PQR server is generating your results - this page will automatically "
#    print "refresh every %s seconds.<P>" % REFRESHTIME
#    print "Thank you for your patience!<P>"
#
#    print "Server Information:<P>"
#    print "<blockquote>"
#    loads = getLoads()
#    if loads != None:
#        print "<font size=2>Server load:</font> <code>%s (1min)  %s (5min)  %s (15min)</code><BR>" % (loads[0], loads[1], loads[2])
#
#    print "<font size=2>Server time:</font> <code>%s</code><BR>" % (time.asctime(time.localtime()))
#    print "</blockquote>"
#
#    print "</BODY></HTML>"

# def printRedirector(name, have_opal):
#    """
#        Prints a page which redirects the user to querystatus.cgi
#    """
#
#    print "Content-type: text/html \n"
#    print "<html>"
#    print "<head>"
#    print "<meta http-equiv=\"Refresh\" content=\"0; url=%squerystatus.cgi?jobid=%s&haveopal=", % (WEBSITE, name)
#    if have_opal:
#        print "true",
#    else:
#        print "false",
#    print "\">"
#    print "</head>"
#    print "</html>"

# def getLoads():
#    """
#        Get the system load information for output and logging
#
#        Returns
#            loads:  A three entry list containing the 1, 5, and
#                    15 minute loads. If the load file is not found,
#                    return None.
#    """
#    if LOADPATH == "": return None
#    try:
#        file = open(LOADPATH, 'rU')
#    except IOError:
#        return None
#
#    line = file.readline()
#    words = string.split(line)
#    loads = words[:3]
#
#    return loads


def createResults(header, input, name, time, missedligands=[]):
    """
    Create the results web page for CGI-based runs

    Parameters
        header: The header of the PQR file (string)
        input:   A flag whether an input file has been created (int)
        tmpdir:  The resulting file directory (string)
        name:    The result file root name, based on local time (string)
        time:    The time taken to run the script (float)
        missedligands: A list of ligand names whose parameters could
                 not be assigned. Optional. (list)
    """
    newheader = string.replace(header, "\n", "<BR>")
    newheader = string.replace(newheader, " ", "&nbsp;")

    filename = "%s%s%s.html" % (INSTALLDIR, TMPDIR, name)
    file = open(filename, "w")

    file.write("<html>\n")
    file.write("<head>\n")
    file.write("<title>PDB2PQR Results</title>\n")
    file.write('<link rel="stylesheet" href="%s" type="text/css">\n' % STYLESHEET)
    file.write("</head>\n")

    file.write("<body>\n")
    file.write("<h2>PDB2PQR Results</h2>\n")
    file.write("<P>\n")
    file.write(
        "Here are the results from PDB2PQR.  The files will be available on the "
    )
    file.write(
        "server for a short period of time if you need to re-access the results.<P>\n"
    )

    file.write('<a href="%s%s%s.pqr">%s.pqr</a><BR>\n' % (WEBSITE, TMPDIR, name, name))
    if input:
        file.write(
            '<a href="%s%s%s.in">%s.in</a><BR>\n' % (WEBSITE, TMPDIR, name, name)
        )
    pkaname = "%s%s%s.propka" % (INSTALLDIR, TMPDIR, name)
    if os.path.isfile(pkaname):
        file.write(
            '<a href="%s%s%s.propka">%s.propka</a><BR>\n'
            % (WEBSITE, TMPDIR, name, name)
        )
    typename = "%s%s%s-typemap.html" % (INSTALLDIR, TMPDIR, name)
    if os.path.isfile(typename):
        file.write(
            '<a href="%s%s%s-typemap.html">%s-typemap.html</a><BR>\n'
            % (WEBSITE, TMPDIR, name, name)
        )
    file.write(
        "<P>The header for your PQR file, including any warnings generated, is:<P>\n"
    )
    file.write("<blockquote><code>\n")
    file.write("%s<P>\n" % newheader)
    file.write("</code></blockquote>\n")
    if missedligands != []:
        file.write("The forcefield that you have selected does not have ")
        file.write(
            "parameters for the following ligands in your PDB file.  Please visit "
        )
        file.write(
            '<a href="http://davapc1.bioch.dundee.ac.uk/programs/prodrg/">PRODRG</a> '
        )
        file.write(
            "to convert these ligands into MOL2 format.  This ligand can the be "
        )
        file.write(
            "parameterized in your PDB2PQR calculation using the PEOE_PB methodology via "
        )
        file.write(
            "the 'Assign charges to the ligand specified in a MOL2 file' checkbox:<P>\n"
        )
        file.write("<blockquote><code>\n")
        for item in missedligands:
            file.write("%s<BR>\n" % item)
        file.write("<P></code></blockquote>\n")
    file.write(
        'If you would like to run PDB2PQR again, please click <a href="%s%s">\n'
        % (WEBSITE, WEBNAME)
    )
    file.write("here</a>.<P>\n")
    file.write(
        'If you would like to run APBS with these results, please click <a href="%s../apbs/index.py?pdb2pqr-id=%s">here</a>.<P>\n'
        % (WEBSITE[:-1], name)
    )
    file.write("<P>Thank you for using the PDB2PQR server!<P>\n")
    file.write(
        '<font size="-1"><P>Total time on server: %.2f seconds</font><P>\n' % time
    )
    file.write(
        '<font size="-1"><CENTER><I>Last Updated %s</I></CENTER></font>\n' % __date__
    )
    file.write("</body>\n")
    file.write("</html>\n")


def createError(name, details):
    """
    Create an error results page for CGI-based runs

    Parameters
        name:    The result file root name, based on local time (string)
        details: The details of the error (string)
    """
    filename = "%s%s%s.html" % (INSTALLDIR, TMPDIR, name)
    file = open(filename, "w")

    file.write("<html>\n")
    file.write("<head>\n")
    file.write("<title>PDB2PQR Error</title>\n")
    file.write('<link rel="stylesheet" href="%s" type="text/css">\n' % STYLESHEET)
    file.write("</head>\n")

    file.write("<body>\n")
    file.write("<h2>PDB2PQR Error</h2>\n")
    file.write("<P>\n")
    file.write("An error occurred when attempting to run PDB2PQR:<P>\n")
    file.write("%s<P>\n" % details)
    file.write(
        "If you believe this error is due to a bug, please contact the server administrator.<BR>\n"
    )
    file.write(
        'If you would like to try running PDB2QR again, please click <a href="%s%s">\n'
        % (WEBSITE, WEBNAME)
    )
    file.write("here</a>.<P>\n")
    file.write(
        '<font size="-1"><CENTER><I>Last Updated %s</I></CENTER></font>\n' % __date__
    )
    file.write("</body>\n")
    file.write("</html>\n")


# def startServer(name):
#    """
#        Start the PDB2PQR server.  This function is necessary so
#        that useful information can be displayed to the user - otherwise
#        nothing would be returned until the complete run finishes.
#
#        Parameters
#            name:    The ID name of the final file to create (string)
#        Returns
#            pqrpath: The complete path to the pqr file (string)
#    """
#    cleanTmpdir()
#    path = SRCPATH
#    tmpdir = TMPDIR
#
#    starttime = time.time()
#    pid = os.fork()
#    if pid: #Parent - Create refreshed HTML pages and exit
#        pid2 = os.fork()
#        if pid2: # print to browser and exit
#            #printAcceptance(name)
#            printRedirector(name,False) # prints initial page, it should redirect to querystatus.cgi - just a blank page with immediate redirect, False is for signifying it's not opal, which is passed on by cgi
#            sys.exit()
#        else: # Thread in charge of refreshing pages
#            home = os.getcwd()
#            os.chdir("/")
#            os.setsid()
#            os.umask(0)
#            os.chdir(home)
#            os.close(1)
#            os.close(2)
#            #endname = "%s%s%s.html" % (SRCPATH, TMPDIR, name)
#            filename = "%s%s%s" % (SRCPATH, TMPDIR, name)
#            #tmpname = "%s%s%s-tmp.html" % (SRCPATH, TMPDIR, name)
#            while 1:
#                if os.path.isfile(endname): # when createResults makes endname, exit loop
#                    #refreshname = "%s%s%s.html" % (WEBSITE, TMPDIR, name)
#                    #reftime = 5 # reduces refresh time, since calculation is done
#                    #printProgress(name, refreshname, reftime, starttime)
#                    file = open(filename, 'w')
#                    file.write('done')
#                    file.close()
#                    break
#                else: # until then, keep writing to refresh page with new load and time
#                    #refreshname = "%s%s%s-tmp.html" % (WEBSITE, TMPDIR, name)
#                    #reftime = REFRESHTIME
#                    #printProgress(name, refreshname, reftime, starttime)
#                    file = open(filename, 'w')
#                    file.write('done')
#                    file.close()
#                time.sleep(REFRESHTIME)
#
#            time.sleep(REFRESHTIME)
#            #os.remove(tmpname)
#            os.remove(filename)
#            sys.exit()
#
#    else: # Child - run PDB2PQR
#        # don't know what this does, or if there's even anything going on here, like running pdb2pqr
#        home = os.getcwd()
#        os.chdir("/")
#        os.setsid()
#        os.umask(0)
#        os.chdir(home)
#        os.close(1)
#        os.close(2)
#        pqrpath = "%s%s%s.pqr" % (path, tmpdir, name)
#        return pqrpath



"""
    Driver for PDB2PQR

    This module takes a PDB file as input and performs optimizations
    before yielding a new PDB-style file as output.

    Ported to Python by Todd Dolinsky (todd@ccb.wustl.edu)
    Washington University in St. Louis

    Parsing utilities provided by Nathan A. Baker (Nathan.Baker@pnl.gov)
    Pacific Northwest National Laboratory

    Copyright (c) 2002-2011, Jens Erik Nielsen, University College Dublin; 
    Nathan A. Baker, Battelle Memorial Institute, Developed at the Pacific 
    Northwest National Laboratory, operated by Battelle Memorial Institute, 
    Pacific Northwest Division for the U.S. Department Energy.; 
    Paul Czodrowski & Gerhard Klebe, University of Marburg.

	All rights reserved.

	Redistribution and use in source and binary forms, with or without modification, 
	are permitted provided that the following conditions are met:

		* Redistributions of source code must retain the above copyright notice, 
		  this list of conditions and the following disclaimer.
		* Redistributions in binary form must reproduce the above copyright notice, 
		  this list of conditions and the following disclaimer in the documentation 
		  and/or other materials provided with the distribution.
        * Neither the names of University College Dublin, Battelle Memorial Institute,
          Pacific Northwest National Laboratory, US Department of Energy, or University
          of Marburg nor the names of its contributors may be used to endorse or promote
          products derived from this software without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
	ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
	IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
	INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
	BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
	DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
	LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE 
	OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
	OF THE POSSIBILITY OF SUCH DAMAGE.

"""

__date__  = "5 April 2010"
__author__ = "Todd Dolinsky, Nathan Baker, Jens Nielsen, Paul Czodrowski, Jan Jensen, Samir Unni, Yong Huang"

import glob
import os
import time
import re
import sys
import string
from StringIO import StringIO
#import tempfile
#from src import pdb

#from src import structures
#from src import routines
#from src import protein
#from src import server
from src.pdb import readPDB
#from src.utilities import *
#from src.structures import *
from src.definitions import Definition
#from src.forcefield import *
#from src.routines import *
from src.protein import Protein
#from src.server import *
#from src.hydrogens import *
from src import utilities
from src.server import setID, createError
from src.aconf import (STYLESHEET, 
                       WEBSITE, 
                       PDB2PQR_OPAL_URL,
                       HAVE_PDB2PQR_OPAL,
                       INSTALLDIR,
                       TMPDIR,
                       MAXATOMS, 
                       PDB2PQR_VERSION)

import traceback

__version__ = PDB2PQR_VERSION


from main import runPDB2PQR

def printHeader(pagetitle,have_opal=None,jobid=None):
    """
        Function to print html headers
    """
    if jobid:
        if have_opal:
            print "Location: querystatus.cgi?jobid=%s&typeofjob=opal\n" % (jobid)
        else:
            print "Location: querystatus.cgi?jobid=%s&typeofjob=local\n" % (jobid)

    #print "Content-type: text/html\n"
    print "<HTML>"
    print "<HEAD>"
    print "\t<TITLE>%s</TITLE>" % pagetitle
    print "\t<link rel=\"stylesheet\" href=\"%s\" type=\"text/css\">\n" % STYLESHEET
    print "</HEAD>"
    return

def redirector(name, weboptions):
    """
        Prints a page which redirects the user to querystatus.cgi and writes starting time to file
    """
    
    redirectWait = 3

    utilities.startLogFile(name, 'pdb2pqr_start_time', str(time.time()))
    
    jobid = int(name)
    
    analiticsDict = weboptions.getOptions()
    
    events = {}
    
    events['submission'] = analiticsDict['pdb']+'|'+str(os.environ["REMOTE_ADDR"])
    del analiticsDict['pdb']
    
    events['titration'] = str(analiticsDict.get('ph'))
    if 'ph' in analiticsDict:
        del analiticsDict['ph']
        
    events['apbsInput'] = str(analiticsDict.get('apbs'))
    del analiticsDict['apbs']
    
    #Clean up selected extensions output
    if 'selectedExtensions' in analiticsDict:
        analiticsDict['selectedExtensions'] = ' '.join(analiticsDict['selectedExtensions'])
    
    options = ','.join(str(k)+':'+str(v) for k,v in analiticsDict.iteritems())
    events['options']=options

    eventsScriptString = ''
    for event in events:
        eventsScriptString += utilities.getEventTrackingString(category='submissionData',
                                                               action=event, 
                                                               label=events[event]) 
        
    redirectURL = "{website}querystatus.cgi?jobid={jobid}&calctype=pdb2pqr".format(website=WEBSITE, 
                                                                                   jobid=jobid)

    string = """
<html>
    <head>
        {trackingscript}
        <script type="text/javascript">
            {trackingevents}
        </script>
        <meta http-equiv="Refresh" content="{wait}; url={redirectURL}"> 
        <link rel="stylesheet" href="{website}pdb2pqr.css"type="text/css">
    </head>
    <body>
    <center>
        You are being automatically redirected to a new location.<br />
        If your browser does not redirect you in {wait} seconds, or you do
        not wish to wait, <a href="{redirectURL}">click here</a></center>. 
    </body>
</html>""".format(trackingscript=utilities.getTrackingScriptString(jobid=jobid), 
                  trackingevents=eventsScriptString, redirectURL=redirectURL, wait=redirectWait, website=WEBSITE)
    return string

def sanitizeFileName(fileName):
    fileName = re.split(r'[/\\]',fileName)[-1]
    fileName = fileName.replace(' ', '_')
    fileName = fileName.replace('-', '_')
    return fileName

class WebOptionsError(Exception):
    pass

class WebOptions(object):
    '''Helper class for gathering and querying options selected by the user'''
    def __init__(self, form):
        '''Gleans all information about the user selected options and uploaded files.
        Also validates the user input. Raises WebOptionsError if there is any problems.'''
        
        #options to pass to runPDB2PQR
        self.runoptions = {}
        #Additional options to pass to google analytics along with the run options.
        #These are included in has_key(), __contains__(), and __getitem__() calls.
        self.otheroptions = {}
        
        self.runoptions['debump'] = form.has_key("DEBUMP")
        self.runoptions['opt'] = form.has_key("OPT")
        
        if form.has_key('FF'):
            self.ff = form["FF"].value.lower()
        else:
            raise WebOptionsError('Force field type missing from form.')
        
        if form.has_key("PDBID") and form["PDBID"].value and form["PDBSOURCE"].value == 'ID':
            self.pdbfile = utilities.getPDBFile(form["PDBID"].value)
            if self.pdbfile is None:
                raise WebOptionsError('The pdb ID provided is invalid.')
            self.pdbfilestring = self.pdbfile.read()
            self.pdbfile = StringIO(self.pdbfilestring)
            self.pdbfilename = form["PDBID"].value
        elif form.has_key("PDB") and form["PDB"].filename and form["PDBSOURCE"].value == 'UPLOAD':
            self.pdbfilestring = form["PDB"].value
            self.pdbfile = StringIO(self.pdbfilestring)
            self.pdbfilename = sanitizeFileName(form["PDB"].filename)
        else:
            raise WebOptionsError('You need to specify a pdb ID or upload a pdb file.')
            
        if form.has_key("PKACALCMETHOD"):
            if form["PKACALCMETHOD"].value != 'none':
                if not form.has_key('PH'):
                    raise WebOptionsError('Please provide a pH value.')
                
                phHelp = 'Please choose a pH between 0.0 and 14.0.'
                try:
                    ph = float(form["PH"].value)
                except ValueError:
                    raise WebOptionsError('The pH value provided must be a number!  ' + phHelp)
                if ph < 0.0 or ph > 14.0: 
                    text = "The entered pH of %.2f is invalid!  " % ph
                    text += phHelp
                    raise WebOptionsError(text)
                self.runoptions['ph'] = ph
                #build propka and pdb2pka options
                if form['PKACALCMETHOD'].value == 'propka':
                    self.runoptions['ph_calc_method'] = 'propka'
                    self.runoptions['ph_calc_options'] = utilities.createPropkaOptions(ph, False)
                if form['PKACALCMETHOD'].value == 'pdb2pka':
                    self.runoptions['ph_calc_method'] = 'pdb2pka'
                    self.runoptions['ph_calc_options'] = {'output_dir': 'pdb2pka_output',
                                                          'clean_output': True,
                                                          'pdie': 8,
                                                          'sdie': 80,
                                                          'pairene': 1.0}
                 
        self.otheroptions['apbs'] = form.has_key("INPUT")
        self.otheroptions['whitespace'] = form.has_key("WHITESPACE")
        
        if self.ff == 'user':
            if form.has_key("USERFF") and form["USERFF"].filename:
                self.userfffilename = sanitizeFileName(form["USERFF"].filename)
                self.userffstring = form["USERFF"].value
                self.runoptions['userff'] = StringIO(form["USERFF"].value)
            else:
                text = "A force field file must be provided if using a user created force field."
                raise WebOptionsError(text)
                
            if form.has_key("USERNAMES") and form["USERNAMES"].filename:
                self.usernamesfilename = sanitizeFileName(form["USERNAMES"].filename)
                self.usernamesstring = form["USERNAMES"].value
                self.runoptions['usernames'] = StringIO(form["USERNAMES"].value)
            else:
                text = "A names file must be provided if using a user created force field."
                raise WebOptionsError(text)
            
        if form.has_key("FFOUT") and form["FFOUT"].value != "internal":
            self.runoptions['ffout'] = form["FFOUT"].value
                
        self.runoptions['chain'] = form.has_key("CHAIN")
        self.runoptions['typemap'] = form.has_key("TYPEMAP")
        self.runoptions['neutraln'] = form.has_key("NEUTRALN")
        self.runoptions['neutralc'] = form.has_key("NEUTRALC")
        self.runoptions['drop_water'] = form.has_key("DROPWATER")
        
        if (self.runoptions['neutraln'] or self.runoptions['neutraln']) and \
            self.ff != 'parse':
            raise WebOptionsError('Neutral N-terminus and C-terminus require the PARSE forcefield.')
        
        if form.has_key("LIGAND") and form['LIGAND'].filename:
            self.ligandfilename=sanitizeFileName(form["LIGAND"].filename)
            ligandfilestring = form["LIGAND"].value
            # for Windows and Mac style newline compatibility for pdb2pka
            ligandfilestring = ligandfilestring.replace('\r\n', '\n')
            self.ligandfilestring = ligandfilestring.replace('\r', '\n')
            
            self.runoptions['ligand'] = StringIO(self.ligandfilestring)
            
        if self.pdbfilename[-4:]==".pdb":
            self.pqrfilename = "%s.pqr" % self.pdbfilename[:-4]
        else:
            self.pqrfilename = "%s.pqr" % self.pdbfilename
            
        #Always turn on summary and verbose.
        self.runoptions['verbose'] = True
        self.runoptions['selectedExtensions'] = ['summary']
        
    def getLoggingList(self):
        '''Returns a list of options the user has turned on.
        Used for logging jobs later in usage.txt'''
        results = []
        
        for key in self:
            if self[key]:
                results.append(key)
                
        return results
        
    def getRunArguments(self):
        '''Returns argument suitable for runPDB2PQR'''
        return self.runoptions.copy()
    
    def getOptions(self):
        '''Returns all options for reporting to Google analytics'''
        options = self.runoptions.copy()
        options.update(self.otheroptions)
        
        options['ff'] = self.ff
        
        options['pdb'] = self.pdbfilename
        
        #propkaOptions is redundant.
        if options.has_key('ph_calc_options'):
            del options['ph_calc_options']
        
        if options.has_key('ligand'):
            options['ligand'] = self.ligandfilename
            
        if options.has_key('userff'):
            options['userff'] = self.userfffilename
            
        if options.has_key('usernames'):
            options['usernames'] = self.usernamesfilename
        
        return options
    
    def getCommandLine(self):
        commandLine = []
        
        if not self.runoptions['debump']:
            commandLine.append('--nodebump')
            
        if not self.runoptions['opt']:
            commandLine.append('--noopt')
            
        if 'ph' in self.runoptions:
            commandLine.append('--with-ph=%s' % self.runoptions['ph'])
            
        if 'ph_calc_method' in self.runoptions:
            commandLine.append('--ph-calc-method=%s' % self.runoptions['ph_calc_method'])
            
        if self.runoptions['drop_water']:
            commandLine.append('--drop-water')
        
        if self.otheroptions['apbs']:
            commandLine.append('--apbs-input')
            
        if self.otheroptions['whitespace']:
            commandLine.append('--whitespace')
            
        if 'userff' in self.runoptions and self.ff == 'user':
            commandLine.append('--userff=%s' % self.userfffilename)
            commandLine.append('--usernames=%s' % self.usernamesfilename)
        else:
            commandLine.append('--ff=%s' % self.ff)
            
        if 'ffout' in self.runoptions:
            commandLine.append('--ffout=%s' % self.runoptions['ffout'])
            
        for o in ('chain', 'typemap', 'neutraln', 'neutralc', 'verbose'):
            if self.runoptions[o]:
                commandLine.append('--' + o)
            
        if 'ligand' in self.runoptions:
            commandLine.append('--ligand=%s' % self.ligandfilename)
            
        for ext in self.runoptions.get('selectedExtensions',[]):
            commandLine.append('--%s' % ext)
            
        commandLine.append(self.pdbfilename)
        
        commandLine.append(self.pqrfilename)
        
        return ' '.join(commandLine)
    
    def __contains__(self, item):
        '''Helper for checking for the presence of an option'''
        return item in self.runoptions or item in self.otheroptions
    
    def has_key(self, item):
        '''Helper for checking for the presence of an option'''
        return item in self.runoptions or item in self.otheroptions
    
    def __iter__(self):
        for key in self.runoptions:
            yield key
        
        for key in self.otheroptions:
            yield key
            
    def __getitem__(self, key):
        return self.runoptions[key] if key in self.runoptions else self.otheroptions[key]

def handleOpal(weboptions):
    '''
        Handle opal based run.
    '''
    # Opal-specific import statements
    from AppService_client import AppServiceLocator, launchJobRequest
    from AppService_types import ns0

    inputFiles = []
    
    if 'userff' in weboptions:
        ffFile = ns0.InputFileType_Def('inputFile')
        ffFile._name = weboptions.userfffilename
        ffFile._contents = weboptions.userffstring
        inputFiles.append(ffFile)
        
    if 'usernames' in weboptions:
        namesFile = ns0.InputFileType_Def('inputFile')
        namesFile._name = weboptions.usernamesfilename
        namesFile._contents = weboptions.usernamesstring
        inputFiles.append(namesFile)
        
    if 'ligand' in weboptions:
        ligandFile = ns0.InputFileType_Def('inputFile')
        ligandFile._name = weboptions.ligandfilename
        ligandFile._contents = weboptions.ligandfilestring
        inputFiles.append(ligandFile)
        
    pdbOpalFile = ns0.InputFileType_Def('inputFile')
    pdbOpalFile._name = weboptions.pdbfilename
    pdbOpalFile._contents = weboptions.pdbfilestring
    inputFiles.append(pdbOpalFile)
     
    # launch job   
    appLocator = AppServiceLocator()
    appServicePort = appLocator.getAppServicePort(PDB2PQR_OPAL_URL)
    
    req = launchJobRequest()
    req._argList = weboptions.getCommandLine()
    req._inputFile=inputFiles

    try:
        resp=appServicePort.launchJob(req)
    except Exception, e:
        printHeader("PDB2PQR Job Submission - Error")
        print "<BODY>\n<P>"
        print "There was an error with your job submission<br>"
        print "</P>"
        print "<script type=\"text/javascript\">"
        print "var gaJsHost = ((\"https:\" == document.location.protocol) ? \"https://ssl.\" : \"http://www.\");"
        print "document.write(unescape(\"%3Cscript src=\'\" + gaJsHost + \"google-analytics.com/ga.js\' type=\'text/javascript\'%3E%3C/script%3E\"));"
        print "</script>"
        print "<script type=\"text/javascript\">"
        print "try {"
        print "var pageTracker = _gat._getTracker(\"UA-11026338-3\");"
        for key in weboptions:
            print "pageTracker._trackPageview(\"/main_cgi/has_%s_%s.html\");" % (key, weboptions[key])
        print "pageTracker._trackPageview();"
        print "} catch(err) {}</script>"
        print "</BODY>"
        print "</HTML>"
        sys.exit(2)
    
    try:
        starttime = time.time()
        name = setID(starttime)
        
        #Some job parameters logging.
        os.makedirs('%s%s%s' % (INSTALLDIR, TMPDIR, name))
        apbsInputFile = open('%s%s%s/apbs_input' % (INSTALLDIR, TMPDIR, name),'w')
        apbsInputFile.write(str(weboptions["apbs"]))
        apbsInputFile.close()
        
        typemapInputFile = open('%s%s%s/typemap' % (INSTALLDIR, TMPDIR, name),'w')
        typemapInputFile.write(str(weboptions["typemap"]))
        typemapInputFile.close()
        
        pdb2pqrOpalJobIDFile = open('%s%s%s/pdb2pqr_opal_job_id' % (INSTALLDIR, TMPDIR, name), 'w')
        pdb2pqrOpalJobIDFile.write(resp._jobID)
        pdb2pqrOpalJobIDFile.close()
        
        print redirector(name, weboptions)
        
        # Recording CGI run information for PDB2PQR Opal
        pdb2pqrOpalLogFile = open('%s%s%s/pdb2pqr_log' % (INSTALLDIR, TMPDIR, name), 'w')
        pdb2pqrOpalLogFile.write(str(weboptions.getOptions())+'\n'+
                                 str(os.environ["REMOTE_ADDR"]))
        pdb2pqrOpalLogFile.close()

    except StandardError, details:
        print details
        createError(name, details)

def handleNonOpal(weboptions):
    """
        Handle non opal run.
    """
 
    pdblist, errlist = readPDB(weboptions.pdbfile)
    
    dummydef = Definition()
    dummyprot = Protein(pdblist, dummydef)
    if len(pdblist) == 0 and len(errlist) == 0:
        text = "Unable to find PDB file - Please make sure this is "
        text += "a valid PDB file ID!"
        #print "Content-type: text/html\n"
        print text
        sys.exit(2)
    elif dummyprot.numAtoms() > MAXATOMS and weboptions["opt"] == True:
        text = "<HTML><HEAD>"
        text += "<TITLE>PDB2PQR Error</title>"
        text += "<link rel=\"stylesheet\" href=\"%s\" type=\"text/css\">" % STYLESHEET
        text += "</HEAD><BODY><H2>PDB2PQR Error</H2><P>"
        text += "Due to server limits, we are currently unable to optimize "
        text += "proteins of greater than %i atoms on the server (PDB2PQR " % MAXATOMS
        text += "found %s atoms in the selected PDB file).  If you " % dummyprot.numAtoms()
        text += "want to forgo optimization please try the server again.<P>"
        text += "Otherwise you may use the standalone version of PDB2PQR that "
        text += "is available from the <a href=\"http://pdb2pqr.sourceforge.net\">"
        text += "PDB2PQR SourceForge project page</a>."
        text += "<script type=\"text/javascript\">"
        text += "var gaJsHost = ((\"https:\" == document.location.protocol) ? \"https://ssl.\" : \"http://www.\");"
        text += "document.write(unescape(\"%3Cscript src=\'\" + gaJsHost + \"google-analytics.com/ga.js\' type=\'text/javascript\'%3E%3C/script%3E\"));"
        text += "</script>"
        text += "<script type=\"text/javascript\">"
        text += "try {"
        text += "var pageTracker = _gat._getTracker(\"UA-11026338-3\");"
        for key in weboptions:
            text += "pageTracker._trackPageview(\"/main_cgi/has_%s_%s.html\");" % (key, weboptions[key])
        text += "pageTracker._trackPageview();"
        text += "} catch(err) {}</script>"
        text += "</BODY></HTML>"
        #print "Content-type: text/html\n"
        print text
        sys.exit(2)

    try:
        starttime = time.time()
        name = setID(starttime)

        #Some job parameters logging.
        os.makedirs('%s%s%s' % (INSTALLDIR, TMPDIR, name))
        apbsInputFile = open('%s%s%s/apbs_input' % (INSTALLDIR, TMPDIR, name),'w')
        apbsInputFile.write(str(weboptions["apbs"]))
        apbsInputFile.close()
        
        typemapInputFile = open('%s%s%s/typemap' % (INSTALLDIR, TMPDIR, name),'w')
        typemapInputFile.write(str(weboptions["typemap"]))
        typemapInputFile.close()

        statusfile = open('%s%s%s/pdb2pqr_status' % (INSTALLDIR, TMPDIR, name), 'w')
        statusfile.write('running')
        statusfile.close()
        
        # Recording CGI run information for PDB2PQR Opal
        pdb2pqrLogFile = open('%s%s%s/pdb2pqr_log' % (INSTALLDIR, TMPDIR, name), 'w')
        pdb2pqrLogFile.write(str(weboptions.getOptions())+'\n'+
                                 str(weboptions.ff)+'\n'+
                                 str(os.environ["REMOTE_ADDR"]))
        pdb2pqrLogFile.close()


        pid = os.fork()
        if pid:
            print redirector(name, weboptions)
            sys.exit()
        else:
            currentdir = os.getcwd()
            os.chdir("/")
            os.setsid()
            os.umask(0)
            os.chdir(currentdir)
            os.close(1) # not sure if these
            os.close(2) # two lines are necessary
            
            pqrpath = '%s%s%s/%s.pqr' % (INSTALLDIR, TMPDIR, name, name)
            
            orig_stdout = sys.stdout
            orig_stderr = sys.stderr
            sys.stdout = open('%s%s%s/pdb2pqr_stdout.txt' % (INSTALLDIR, TMPDIR, name), 'w')
            sys.stderr = open('%s%s%s/pdb2pqr_stderr.txt' % (INSTALLDIR, TMPDIR, name), 'w')
            
            run_arguements = weboptions.getRunArguments()
            if weboptions.runoptions.get('ph_calc_method', '') == 'pdb2pka':
                run_arguements['ph_calc_options']['output_dir']='%s%s%s/pdb2pka_output' % (INSTALLDIR, TMPDIR, name)
            
            
            header, lines, missedligands = runPDB2PQR(pdblist, 
                                                      weboptions.ff,
                                                      outname = pqrpath,
                                                      commandLine = weboptions.getCommandLine(),
                                                      **weboptions.getRunArguments())
            
            sys.stdout.close()
            sys.stderr.close()
            sys.stdout = orig_stdout
            sys.stderr = orig_stderr

            endtimefile = open('%s%s%s/pdb2pqr_end_time' % (INSTALLDIR, TMPDIR, name), 'w')
            endtimefile.write(str(time.time()))
            endtimefile.close()

            pqrfile = open(pqrpath, "w")
            pqrfile.write(header)
            
            whitespace = weboptions.otheroptions['whitespace']
            for line in lines:
                # Adding whitespaces if --whitespace is in the weboptions
                if whitespace: 
                    if line[0:4] == 'ATOM':
                        newline = line[0:6] + ' ' + line[6:16] + ' ' + line[16:38] + ' ' + line[38:46] + ' ' + line[46:]
                        pqrfile.write("%s\n" % string.strip(newline))
                    elif line[0:6] == 'HETATM':
                        newline = line[0:6] + ' ' + line[6:16] + ' ' + line[16:38] + ' ' + line[38:46] + ' ' + line[46:]
                        pqrfile.write("%s\n" % string.strip(newline))
                else: 
                    pqrfile.write("%s\n" % string.strip(line))
            pqrfile.close()
                    
            if weboptions.otheroptions['apbs']:
                from src import inputgen
                from src import psize
                method = "mg-auto"
                size = psize.Psize()
                size.parseInput(pqrpath)
                size.runPsize(pqrpath)
                async = 0 # No async files here!
                myinput = inputgen.Input(pqrpath, size, method, async, potdx=True)
                myinput.printInputFiles()
                myinput.dumpPickle()
                        
            endtime = time.time() - starttime
            #createResults(header, input, name, endtime, missedligands)
            #logRun(weboptions, endtime, len(lines), weboptions.ff, os.environ["REMOTE_ADDR"])
            #printHeader("PDB2PQR Job Submission",have_opal,jobid=name)
            if 'ligand' in weboptions:
                outputligandfile = open('%s%s%s/%s.mol2' % (INSTALLDIR,TMPDIR, name, name),'w')
                outputligandfile.write(weboptions.ligandfilestring)
                outputligandfile.close()
            outputpdbfile = open('%s%s%s/%s.pdb' % (INSTALLDIR,TMPDIR,name,name),'w')
            outputpdbfile.write(weboptions.pdbfilestring)
            outputpdbfile.close()

            statusfile = open('%s%s%s/pdb2pqr_status' % (INSTALLDIR, TMPDIR, name), 'w')
            statusfile.write('complete\n')
            filelist = glob.glob('%s%s%s/%s*' % (INSTALLDIR, TMPDIR, name, name))
            for filename in filelist:
                statusfile.write(filename+'\n')
            statusfile.close()

    #TODO: Better error reporting.
    #Also, get forked job to properly write error status on failure.
    except StandardError, details:
    #except StandardError as details:
        print traceback.format_exc()
        print sys.exc_info()[0]
        #print details
        createError(name, details)

def mainCGI():
    """
        Main driver for running PDB2PQR from a web page
    """
    print "Content-type: text/html\n"
    import cgi
    import cgitb

    cgitb.enable()
    form = cgi.FieldStorage()
    
    try:
        weboptions = WebOptions(form)
    except WebOptionsError as error:
        print(error)
        sys.exit(2)
        
    if HAVE_PDB2PQR_OPAL:
        handleOpal(weboptions)
    else:
        handleNonOpal(weboptions)
        
    return
    


"""
    Chi extension

    Print the backbone chi angle for each residue in the structure.
    Chi angle is determined by the coordinates of the N, CA, CB (if
    available), and CG/OG/SG atoms (if available).

    Author:  Todd Dolinsky
"""

__date__ = "17 February 2006"
__author__ = "Todd Dolinsky"

from src.utilities import getDihedral


def usage():
    return "Print the per-residue backbone chi angle to {output-path}.chi"


def run_extension(routines, outroot, options):
    """
    Print the list of psi angles

    Parameters
        routines:  A link to the routines object
        outroot:   The root of the output name
        options:   options object
    """

    outname = outroot + ".chi"
    outfile = open(outname, "w")

    routines.write("\nPrinting chi angles for each residue...\n")
    routines.write("Residue     chi\n")
    routines.write("----------------\n")

    # Initialize some variables

    protein = routines.protein

    for residue in protein.getResidues():
        if residue.hasAtom("N"):
            ncoords = residue.getAtom("N").getCoords()
        else:
            continue

        if residue.hasAtom("CA"):
            cacoords = residue.getAtom("CA").getCoords()
        else:
            continue

        if residue.hasAtom("CB"):
            cbcoords = residue.getAtom("CB").getCoords()
        else:
            continue

        if residue.hasAtom("CG"):
            gcoords = residue.getAtom("CG").getCoords()
        elif residue.hasAtom("OG"):
            gcoords = residue.getAtom("OG").getCoords()
        elif residue.hasAtom("SG"):
            gcoords = residue.getAtom("SG").getCoords()
        else:
            continue

        chi = getDihedral(ncoords, cacoords, cbcoords, gcoords)
        routines.write("%s\t%.4f\n" % (residue, chi))
        outfile.write("%s\t%.4f\n" % (residue, chi))

    routines.write("\n")
    outfile.close()



import distutils.sysconfig

import os
from defaults import *
import atexit

Export("codePath")

config_file = "build_config.py"

gcv = distutils.sysconfig.get_config_var

vars = Variables([".variables.cache", config_file], ARGUMENTS)

vars.Add(
    PathVariable("PREFIX", "Install directory", defaultPrefix, PathVariable.PathAccept)
)

vars.Add("URL", "Sets the url of pdb2pqr's web interface.", defaultURL)

vars.Add("OPAL", "Sets the url of a pdb2pqr opal service.", "")
vars.Add("APBS_OPAL", "Sets the url of a apbs opal service.", "")

vars.Add(
    PathVariable(
        "APBS", "Location of the APBS binary if installed", "", PathVariable.PathAccept
    )
)

vars.Add(
    "MAX_ATOMS",
    "Sets the maximum number of atoms in a protein for non-Opal job submission. "
    "Only affects web tools",
    10000,
    None,
    int,
)

vars.Add(
    BoolVariable(
        "BUILD_PDB2PKA",
        "Set to False to skip building ligand and pdb2pka support. Requires numpy.",
        True,
    )
)

vars.Add(
    BoolVariable(
        "REBUILD_SWIG",
        "Set to True to rebuild the swig bindings. Requires swig on the the user path.",
        False,
    )
)

vars.Add(
    BoolVariable(
        "DEBUG", "Set to True to compiled components with debug headers.", False
    )
)

vars.Add("EXTRA_CXXFLAGS", "Set to add extra CXX flags to the build.", "")
vars.Add("EXTRA_LINKFLAGS", "Set to add extra link flags to the build.", "")

# TODO: setup rebuilding of docs.
# THIS SHOULD BE A TARGET!
# AddOption('--rebuild-docs',
#        dest='docs',
#        action='store_true',
#        default=False,
#        help='Rebuild pydocs.')

# Windows: make sure we use correct target arch.
target_arch = "x86_64"
import platform

arch = platform.architecture()
bit_str = arch[0]
if bit_str == "32bit":
    target_arch = "x86"

if os.name == "nt":
    tool_chain = ["default", "mssdk"]
else:
    tool_chain = ["default"]

tool_chain.append("swig")

env = Environment(
    variables=vars,
    MSVC_VERSION="9.0",
    TARGET_ARCH=target_arch,
    MSSDK_VERSION="6.0A",
    tools=tool_chain,
    SWIGFLAGS=["-python", "-c++"],
    SHLIBPREFIX="",
    SHLIBSUFFIX=gcv("SO"),
    LDMODULESUFFIX=gcv("SO"),
)


python_lib = "python" + gcv("VERSION")
env.Append(LIBS=[python_lib])
# To get swig to work on windows.
# env.Append(ENV={'PATH' : os.environ['PATH']})

if os.name == "nt" and "icc" not in tool_chain:
    env.Append(CXXFLAGS=["/EHsc"])

if env["DEBUG"]:
    if os.name == "nt" and "icc" not in tool_chain:
        env.Append(CXXFLAGS=["/DEBUG"])
    else:
        env.MergeFlags("-g")

if env["EXTRA_CXXFLAGS"]:
    env.Append(CXXFLAGS=[env["EXTRA_CXXFLAGS"]])

if env["EXTRA_LINKFLAGS"]:
    env.Append(LINKFLAGS=[env["EXTRA_LINKFLAGS"]])

if os.name == "nt":
    # Check to see if we are running in a virtualenv
    if hasattr(sys, "real_prefix"):
        python_root = sys.real_prefix
    else:
        python_root = sys.prefix
    python_include = os.path.join(python_root, "include")
    python_libs = os.path.join(python_root, "libs")
    env.Append(LIBPATH=[python_libs])
else:
    env.Append(LIBPATH=[gcv("LIBDIR")])

Export("env")

prefix = env["PREFIX"]
prefix = prefix.replace("\\", "/")
if not prefix.endswith("/"):
    prefix += "/"

env["PREFIX"] = prefix

Help(vars.GenerateHelpText(env))

vars.Save(".variables.cache", env)

# Not the optimal way to do this...
# Should figure out how to do it with a delete command
Clean("pdb2pqr.py", ".variables.cache")

url = env["URL"]
# Not sure if this is needed.
if url is not None:
    if not url.endswith("/"):
        url += "/"
    submitAction = url + "pdb2pqr.cgi"
else:
    url = defaultURL
    # Can it always just be this?
    submitAction = "pdb2pqr.cgi"

maxatomsStr = str(env["MAX_ATOMS"])

replacementDict = {
    "@WHICHPYTHON@": pythonBin,
    "@INSTALLDIR@": prefix,
    "@MAXATOMS@": maxatomsStr,
    "@website@": url,
    "@srcpath@": codePath,
    "@PDB2PQR_VERSION@": productVersion,
    "@action@": submitAction,
    "@APBS_LOCATION@": env["APBS"],
    "@APBS_OPAL_URL@": env["APBS_OPAL"],
    "@PDB2PQR_OPAL_URL@": env["OPAL"],
}

# If any replacement strings change recompile those files.
# As the product version can be based on the time this may
# rebuild string replacement files after less than one minute between builds
settingsValues = env.Value(replacementDict)

# We have a separate dict for server.html.in as we need to use regex
# Regex does not play nice with some  possible user strings
# Set up regex to alternately clear tags or wipe sections
if env["OPAL"] == "":
    # Not using opal for pdb2pqr.
    print "not using opal", env["OPAL"]
    withOpalRegex = r"@WITHOPAL@.*?@WITHOPAL@"
    withoutOpalRegex = "@WITHOUTOPAL@"

else:
    # Using opal for pdb2pqr.
    print "using opal", env["OPAL"]
    withOpalRegex = "@WITHOPAL@"
    withoutOpalRegex = r"@WITHOUTOPAL@.*?@WITHOUTOPAL@"


serverHtmlDict = {
    "@website@": url,
    "@PDB2PQR_VERSION@": productVersion,
    "@MAXATOMS@": maxatomsStr,
    "@action@": submitAction,
    withOpalRegex: "",
    withoutOpalRegex: "",
}

chmodAction = Chmod("$TARGET", 0755)
serverHtmlCopySub = CopySub("$TARGET", "$SOURCE", serverHtmlDict, useRegex=True)
normalCopySub = CopySub("$TARGET", "$SOURCE", replacementDict, useRegex=False)

subFiles = [
    ("pdb2pqr.py", "pdb2pqr.py.in", True),
    ("apbs_cgi.cgi", "apbs_cgi.py", True),
    ("visualize.cgi", "visualize.py", True),
    ("querystatus.cgi", "querystatus.py", True),
    ("src/aconf.py", "src/aconf.py.in", False),
    ("html/server.html", "html/server.html.in", False),
    ("3dmol/js/3dmol.js", "3dmol/js/3dmol.js.in", False),
]

compile_targets = []

for target, source, chmod in subFiles:
    actions = [normalCopySub] if target != "html/server.html" else [serverHtmlCopySub]
    if chmod:
        actions.append(chmodAction)
    result = env.Command(target, source, actions)
    compile_targets.append(result)
    if target == "pdb2pqr.py":
        pdb2pqr = result
        Export("pdb2pqr")
    Default(result)
    Depends(result, settingsValues)

Default(env.Command("pdb2pqr.cgi", "pdb2pqr.py", Copy("$TARGET", "$SOURCE")))

# Check to see why we can't build pdb2pka.
numpy_error = False
if env["BUILD_PDB2PKA"]:
    try:
        import numpy
    except ImportError:
        print "WARNING: PDB2PKA build skipped, numpy not installed. Ligand support will not be available."
        numpy_error = True

    if not numpy_error:
        import distutils

        env.Append(CPPPATH=[distutils.sysconfig.get_python_inc(), numpy.get_include()])

        alg_srcs = ["pdb2pka/substruct/Algorithms.cpp"]

        algorithms_pyc = env.LoadableModule(
            "pdb2pka/substruct/Algorithms", ["pdb2pka/substruct/Algorithms.cpp"]
        )

        Default(algorithms_pyc)
        Alias("algorithms", algorithms_pyc)
        compile_targets.append(algorithms_pyc)

        if os.name == "nt":
            alg_msvs_env = env.Clone(
                MSVSBUILDCOM="cd .. && "
                + pythonBin
                + " scons/scons.py algorithms DEBUG=True",
                MSVSCLEANCOM="cd .. && " + pythonBin + " scons/scons.py -c algorithms",
                MSVSREBUILDCOM="cd .. && "
                + pythonBin
                + " scons/scons.py -c algorithms && "
                + pythonBin
                + " scons/scons.py algorithms DEBUG=True",
            )

            algorithms_project = alg_msvs_env.MSVSProject(
                target="msvs/Algorithms" + env["MSVSPROJECTSUFFIX"],
                auto_build_solution=0,
                srcs=["../pdb2pka/substruct/Algorithms.cpp"],
                buildtarget=algorithms_pyc[0],
                variant="Debug|x64",
            )
            Alias("msvs", algorithms_project)

        pmc_srcs = ["pdb2pka/pMC_mult.cpp", "pdb2pka/pMC_mult_wrap.cpp"]

        if env["REBUILD_SWIG"]:
            pmc_pyc = env.LoadableModule(
                "pdb2pka/_pMC_mult", ["pdb2pka/pMC_mult.cpp", "pdb2pka/pMC_mult.i"]
            )
            build_swig = " REBUILD_SWIG=True"
        else:
            pmc_pyc = env.LoadableModule(
                "pdb2pka/_pMC_mult",
                ["pdb2pka/pMC_mult.cpp", "pdb2pka/pMC_mult_wrap.cpp"],
            )
            build_swig = " "

        Default(pmc_pyc)
        Alias("pmc_mult", pmc_pyc)
        compile_targets.append(pmc_pyc)

        if os.name == "nt":
            pmc_msvs_env = env.Clone(
                MSVSBUILDCOM="cd .. && "
                + pythonBin
                + " scons/scons.py pmc_mult DEBUG=True"
                + build_swig,
                MSVSCLEANCOM="cd .. && " + pythonBin + " scons/scons.py -c pmc_mult",
                MSVSREBUILDCOM="cd .. && "
                + pythonBin
                + " scons/scons.py -c pmc_mult && "
                + pythonBin
                + " scons/scons.py pmc_mult DEBUG=True"
                + build_swig,
            )

            pmc_mult_project = pmc_msvs_env.MSVSProject(
                target="msvs/pMC_mult" + env["MSVSPROJECTSUFFIX"],
                auto_build_solution=0,
                srcs=["../pdb2pka/pMC_mult.cpp", "../pdb2pka/pMC_mult_wrap.cpp"],
                incs=["../pdb2pka/pMC_mult.h"],
                resources=["../pdb2pka/pMC_mult.i"],
                buildtarget=pmc_pyc[0],
                variant="Debug|x64",
            )
            Alias("msvs", pmc_mult_project)

            pdb2pqr_sln = env.MSVSSolution(
                target="msvs/pdb2pqr" + env["MSVSSOLUTIONSUFFIX"],
                projects=[pmc_mult_project, algorithms_project],
                variant="Debug|x64",
            )

            Alias("msvs", pdb2pqr_sln)

SConscript("tests/SConscript")

SConscript("SConscript-install.py", exports="env compile_targets")

SConscript("SConscript-error.py")


def print_default_message(target_list):
    target_list = map(str, target_list)
    if any("test" in x for x in target_list):
        return
    if GetOption("clean"):
        return
    if not GetOption("help"):

        print
        print "TARGETS:", target_list
        print
        print "========================"
        print "Configuration Parameters"
        print "========================"
        print
        print "Version:", productVersion
        print "Install directory:", env["PREFIX"]
        if numpy_error:
            print
            print "WARNING: PDB2PKA build skipped, numpy not installed. Ligand support will not be available."
            print
        else:
            print "pdb2pka and ligand support:", env["BUILD_PDB2PKA"]
        print "Path to the website directory:", url
        if env["OPAL"] == "":
            print "PDB2PQR jobs run via the web interface will be forked on the server."
        else:
            print "PDB2PQR jobs run via the web interface will be run via opal at", env[
                "OPAL"
            ]
    else:
        print
        print 'Run "python scons/scons.py" to build pdb2pqr.'

    print
    print "The preferred way to configure the build is by editing the file", config_file
    print
    print "Run scons with the python that you intend to use with pdb2pqr."
    print 'For example: "/opt/bin/python scons/scons.py" will setup pdb2pqr to be run with /opt/bin/python'

    if "install" not in target_list:
        print
        print 'Run "python scons/scons.py install" to install pdb2pqr in', env["PREFIX"]

    print
    print 'Run "python scons/scons.py basic-test" for a basic functionality test'
    print 'Run "python scons/scons.py advanced-test" for a single test of ligand and PROPKA support. Requires numpy and PDB2PKA support compiled.'
    print 'Run "python scons/scons.py complete-test" for a complete test of all functionality EXCEPT PDB2PKA. Requires numpy and PDB2PKA support compiled.'
    print 'Run "python scons/scons.py pdb2pka-test" for a test of PDB2PKA functionality.'
    print "    Requires numpy, PDB2PKA support compiled AND the APBS python libraries compiled and installed in the pdb2pka directory."

    print
    print "To setup a web service create a symbolic link to", env[
        "PREFIX"
    ], "that enables you to view", env["URL"], 'after running "scons/scons.py install"'
    print
    print 'Run "python scons/scons.py msvs" to build Visual Studio projects for the Algorithms and pMC_mult modules.'
    print "VS project generation is not well supported in scons. Resulting projects should build using NMAKE but cannot be used for debugging."
    print "The resulting projects will need to modified to use VS natively to compile the code or debug."

    if "install" in target_list:
        print
        print "pdb2pqr installed in", env["PREFIX"]


atexit.register(print_default_message, BUILD_TARGETS)



"""
    Hbond extension

    Find all hydrogen bonds as determined by the cutoffs specified.
    Uses PDB2PQR to determine donors and acceptors, and displays
    all available bonds to file. 
    
    The original bonding parameters were an angle of 20.0, 
    distance of 3.30, and using the old method for calculating 
    distance.
    
    The original parameters for WHAT-IF output was an angle of
    90.0, distance of 3.30, and using the old method for 
    calculating distance. 

    Authors:  Todd Dolinsky, Michael J Bradley, Julie Mitchell, and Kyle Monson
"""

__date__ = "17 February 2006"
__author__ = "Todd Dolinsky, Michael J Bradley, Julie Mitchell, and Kyle Monson"
# NOTE: This extension edited and updated on 05 August 2008 by Michael J Bradley
# and again on 17 October by Kyle Monson.

# NOTE: The current defaults for hbonds used below were utilized in the following study:
# Bradley MJ, Chivers PT, Baker NA. Molecular dynamics simulation of the
# Escherichia coli NikR protein: Equilibrium conformational fluctuations reveal
# inter-domain allosteric communication pathways. Journal of Molecular Biology, 378,
# 1155-1173, 2008.  http://dx.doi.org/10.1016/j.jmb.2008.03.010

from src.utilities import distance, getAngle
from src.routines import Cells
from math import cos
import extensions

ANGLE_CUTOFF = 30.0  # A - D - H(D) angle
DIST_CUTOFF = 3.4  # D to A distance


def addExtensionOptions(extensionGroup):
    """
    Add options to set output type, angle cutoff, distance cutoff, and distance calculating method.
    """
    extensionGroup.add_option(
        "--whatif",
        dest="whatif",
        action="store_true",
        default=False,
        help="Change hbond output to WHAT-IF format.",
    )

    extensionGroup.add_option(
        "--angle_cutoff",
        dest="angle_cutoff",
        type="float",
        action="store",
        default=ANGLE_CUTOFF,
        help="Angle cutoff to use when creating hbond data (default %s)" % ANGLE_CUTOFF,
    )

    extensionGroup.add_option(
        "--distance_cutoff",
        dest="distance_cutoff",
        type="float",
        action="store",
        default=DIST_CUTOFF,
        help="Distance cutoff to use when creating hbond data (default %s)"
        % DIST_CUTOFF,
    )

    extensionGroup.add_option(
        "--old_distance_method",
        dest="old_distance_method",
        action="store_true",
        default=False,
        help="Use distance from donor hydrogen to acceptor to calculate distance used with --distance_cutoff.",
    )


def usage():
    return "Print a list of hydrogen bonds to {output-path}.hbond"


# TODO: replace this with a ''.format() call.
def _residueString(residue, name):
    return "%4d %-4s (%4d  ) %s     %-4s" % (
        residue.resSeq,
        residue.name,
        residue.resSeq,
        residue.chainID,
        name,
    )


def create_hbond_output(
    routines,
    outfile,
    whatif=False,
    angleCutoff=ANGLE_CUTOFF,
    distanceCutoff=DIST_CUTOFF,
    oldDistanceMethod=False,
):

    routines.write("Printing hydrogen bond list...\n")

    output = extensions.extOutputHelper(routines, outfile)

    cellsize = int(distanceCutoff + 1.0 + 1.0)
    protein = routines.protein
    routines.setDonorsAndAcceptors()
    routines.cells = Cells(cellsize)
    routines.cells.assignCells(protein)

    for donor in protein.getAtoms():

        # Grab the list of donors
        if not donor.hdonor:
            continue
        donorhs = []
        for bond in donor.bonds:
            if bond.isHydrogen():
                donorhs.append(bond)
        if donorhs == []:
            continue

        # For each donor, grab all acceptors
        closeatoms = routines.cells.getNearCells(donor)
        for acc in closeatoms:
            if not acc.hacceptor:
                continue
            if donor.residue == acc.residue:
                continue

            # TODO: do we need to do this for plain hbond stuff?
            if whatif and (donor.residue.chainID == acc.residue.chainID):
                continue

            # Do new style distance check
            if not oldDistanceMethod:
                dist = distance(donor.getCoords(), acc.getCoords())
                if dist > distanceCutoff:
                    continue

            for donorh in donorhs:

                # Do old style distance check
                if oldDistanceMethod:
                    dist = distance(donorh.getCoords(), acc.getCoords())
                    if dist > distanceCutoff:
                        continue

                # Do angle check
                angle = getAngle(acc.getCoords(), donor.getCoords(), donorh.getCoords())
                if angle > angleCutoff:
                    continue

                if whatif:
                    if donor.tempFactor > 60.0:
                        continue
                    if acc.tempFactor > 60.0:
                        continue

                    thisBstring = "B" if donor.isBackbone() else "S"
                    thatBstring = "B" if acc.isBackbone() else "S"

                    score = (1.7 / dist) * cos(angle * 3.142 / 180.0)
                    output.write(_residueString(donor.residue, donor.name))
                    output.write("-> ")
                    output.write(_residueString(acc.residue, acc.name))
                    output.write(
                        "Sym=   1 Val= %6.3lf  DA=%6.2f  DHA=%6.2f (%s-%s)\n"
                        % (score, dist, angle, thisBstring, thatBstring)
                    )
                #                    outfile.write("%4d %-4s (%4d  ) %s     %-4s-> %4d %-4s (%4d  ) %s     %-4sSym=   1 Val= %6.3lf  DA=%6.2f  DHA=%6.2f (%s-%s)\n" % \
                #                      (donor.residue.resSeq,donor.residue.name,donor.residue.resSeq, donor.residue.chainID,donor.name,acc.residue.resSeq,acc.residue.name,acc.residue.resSeq, acc.residue.chainID,acc.name, score, dist, angle, thisBstring, thatBstring))

                else:
                    s = "Donor: %s %s\tAcceptor: %s %s\tdist: %.2f\tAngle: %.2f\n" % (
                        donor.residue,
                        donor.name,
                        acc.residue,
                        acc.name,
                        dist,
                        angle,
                    )
                    output.write(s)

    routines.write("\n")


def run_extension(routines, outroot, options):
    """
    Print a list of hydrogen bonds.

    Parameters
        routines:  A link to the routines object
        outroot:   The root of the output name
        options:   options object
    """

    outname = outroot + ".hbond"
    with open(outname, "w") as outfile:
        create_hbond_output(
            routines,
            outfile,
            whatif=options.whatif,
            angleCutoff=options.angle_cutoff,
            distanceCutoff=options.distance_cutoff,
            oldDistanceMethod=options.old_distance_method,
        )



"""
    Forcefield.py

    This module takes a pdblist as input and replaces the occupancy and
    tempfactor fields with charge and radius fields, with values as defined
    by a particular forcefield.  The forcefield structure is modeled off of
    the structures.py file, where each forcefield is considered a chain of
    residues of atoms.

    ----------------------------
   
    PDB2PQR -- An automated pipeline for the setup, execution, and analysis of
    Poisson-Boltzmann electrostatics calculations

    Copyright (c) 2002-2011, Jens Erik Nielsen, University College Dublin; 
    Nathan A. Baker, Battelle Memorial Institute, Developed at the Pacific 
    Northwest National Laboratory, operated by Battelle Memorial Institute, 
    Pacific Northwest Division for the U.S. Department Energy.; 
    Paul Czodrowski & Gerhard Klebe, University of Marburg.

	All rights reserved.

	Redistribution and use in source and binary forms, with or without modification, 
	are permitted provided that the following conditions are met:

		* Redistributions of source code must retain the above copyright notice, 
		  this list of conditions and the following disclaimer.
		* Redistributions in binary form must reproduce the above copyright notice, 
		  this list of conditions and the following disclaimer in the documentation 
		  and/or other materials provided with the distribution.
        * Neither the names of University College Dublin, Battelle Memorial Institute,
          Pacific Northwest National Laboratory, US Department of Energy, or University
          of Marburg nor the names of its contributors may be used to endorse or promote
          products derived from this software without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
	ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
	IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
	INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
	BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
	DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
	LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE 
	OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
	OF THE POSSIBILITY OF SUCH DAMAGE.

    ----------------------------
"""

__date__ = "6 November 2007"
__author__ = "Todd Dolinsky, Yong Huang"

import string
import sys
import getopt
import os
import re

from xml import sax
from utilities import *
from errors import PDBInputError, PDBInternalError

class ForcefieldHandler(sax.ContentHandler):
   
    def __init__(self, map, reference):
        self.oldresname = None
        self.oldatomname = None
        self.curelement = None
        self.newatomname = None
        self.newresname = None
        self.atommap = {}
        self.map = map
        self.reference = reference           

    def updateMap(self, toname, fromname, map):
        """
            Update the given map by adding a pointer from a new
            name to an object.

            Parameters
                toname:  The new name for the object (string)
                fromname:  The old name for the object (string)
                map:  A dictionary of items (dict)
        """
        fromobj = map[fromname]
        if isinstance(fromobj, ForcefieldResidue):
            if toname not in map:
                newres = ForcefieldResidue(fromname)
                map[toname] = newres
            for atomname in fromobj.atoms:
                map[toname].atoms[atomname] = fromobj.atoms[atomname]
        elif isinstance(fromobj, ForcefieldAtom):
            map[toname] = fromobj
            
                        
    def findMatchingNames(self, regname, map):
        """
            Find a list of strings that match the given regular
            expression.

            Parameters
                regname: The regular expression (string)
                map:  The dictionary to search (dict)
                
            Returns
                list:  A list of regular expression objects that match
                       the regular expression.
        """
        list = [] 
        regname += "$"
     
        # Find the existing items that match this string

        for name in map:
            regexp = re.compile(regname).match(name)
            if regexp:
                list.append(regexp)

        return list
   
    def startElement(self, name, attributes):
        """
            Override the startElement function to keep track of the current
            element.
        """
        if name != "name": self.curelement = name
            
    def endElement(self, name):
        """
            At the end of the element, act on the stored information.

            Parameters
                name:  The name of the element (string)
        """
        if name == "residue":
            if self.oldresname != None:  # Make a new residue hook
              
                newreslist = self.findMatchingNames(self.newresname, self.reference)
                if self.oldresname.find("$group") >= 0:  # Multiple new residues
                    for resitem in newreslist:
                        resname = resitem.string
                        group = resitem.group(1)
                        fromname = self.oldresname.replace("$group", group)
                        if fromname in self.map:
                            self.updateMap(resname, fromname, self.map)
                              
                else: # Work with a single new residue name
                    oldreslist = self.findMatchingNames(self.oldresname, self.map)
                    for resitem in newreslist:
                        resname = resitem.string
                        self.updateMap(resname, self.oldresname, self.map)

            # If this was only a residue conversion, exit
                
            if self.atommap == {}:
                self.oldresname = None
                self.newresname = None
                return

            # Apply atom conversions for all appropriate residues
           
            resmatchlist = self.findMatchingNames(self.newresname, self.map)
            for resitem in resmatchlist:
                residue = self.map[resitem.string]        
                for newname in self.atommap:        
                    oldname = self.atommap[newname]
                    if oldname not in residue.atoms: continue
                    self.updateMap(newname, oldname, residue.atoms)
                    
            # Clean up

            self.oldresname = None
            self.newresname = None
            self.atommap = {}
            
        elif name == "atom": 

            self.atommap[self.newatomname] = self.oldatomname
            self.oldatomname = None
            self.newatomname = None

        else: # Just free the current element namespace
            self.curelement = ""

        return self.map

    def characters(self, text):
        """
            Store the information in the object for future use/

            Parameters
                text:  The text value between the XML tags
        """
        if text.isspace(): return
        text = str(text)   
        if self.curelement == "residue":
            self.newresname = text       
        elif self.curelement == "atom":
            self.newatomname = text      
        elif self.curelement == "useatomname":
            self.oldatomname = text    
        elif self.curelement == "useresname":
            self.oldresname = text
     
class Forcefield:
    """
        Forcefield class

        The forcefield class contains definitions for a given forcefield.
        Each forcefield object contains a dictionary of residues, with each
        residue containing a dictionary of atoms.  Dictionaries are used
        instead of lists as the ordering is not important. The forcefield
        definition files are unedited, directly from the forcefield - all
        transformations are done within.

    """
    #TODO: pass ff and ff names file like objects instead of sorting out whether
    #to use user created files here. 
    def __init__(self, ff, definition, userff, usernames = None):
        """
            Initialize the class by parsing the definition file

            Parameters
                ff: The name of the forcefield (string) can be None.
                definition: The definition objects
                userff:  A link to the file for CGI based user-defined
                         forcefields
        """
        self.map = {}
        self.name = str(ff)
        defpath = ""

        if userff == None:
            defpath = getFFfile(ff)
            if defpath == "":
                raise PDBInputError, "Unable to find forcefield parameter file %s!" % self.name
          
            file = open(defpath, 'rU')

        else: 
            file = userff

        lines = file.readlines()
        for line in lines:
            if not line.startswith("#"):
                fields = string.split(line)
                if fields == []: continue  
                try:
                    resname = fields[0]
                    atomname = fields[1]
                    charge = float(fields[2])
                    radius = float(fields[3])
                except ValueError:
                    txt = "Unable to recognize user-defined forcefield file" 
                    if defpath != "": txt += " %s!" % defpath
                    else: txt += "!"
                    txt += " Please use a valid parameter file."
                    raise PDBInputError(txt)
            
                try:
                    group = fields[4]
                    atom = ForcefieldAtom(atomname, charge, radius, resname, group)
                except IndexError:
                    atom = ForcefieldAtom(atomname, charge, radius, resname)

                myResidue = self.getResidue(resname)
                if myResidue == None:
                    myResidue = ForcefieldResidue(resname)
                    self.map[resname] = myResidue
                myResidue.addAtom(atom)

        file.close()

        # Now parse the XML file, associating with FF objects -
        # This is not necessary (if canonical names match ff names)
        defpath = getNamesFile(ff)
        if defpath != "":            
            handler = ForcefieldHandler(self.map, definition.map)
            sax.make_parser()            
            if usernames != None:
                namesfile = usernames
                sax.parseString(namesfile.read(), handler)                
            else:
                namesfile = open(defpath)
                sax.parseString(namesfile.read(), handler)
            namesfile.close()
        else: 
            handler = ForcefieldHandler(self.map, definition.map)
            sax.make_parser()

            if usernames != None:
                namesfile = usernames            
                sax.parseString(namesfile.read(), handler)
            else:
                raise PDBInputError("Please provide a valid .names file!")
            namesfile.close()


    def hasResidue(self, resname):
        """
            Check if the residue name is in the map or not.

            Parameters
                resname:  The name to search for (string)

            Returns
                1 if the resname is in the map, 0 otherwise.
        """
        if resname in self.map: return 1
        else: return 0

    def getResidue(self, resname):
        """
            Return the residue object with the given resname

            Parameters
                resname: The name of the residue (string)
            Returns
                residue: The residue object (ForcefieldResidue)
        """
        if self.hasResidue(resname): return self.map[resname]
        else: return None

   
    def getNames(self, resname, atomname):
        """
            Get the actual names associated with the input fields.
            The names passed in point to ForcefieldResidue and
            ForcefieldAtom objects which may have different names;
            grab these names and return.

            Parameters
                resname:  The residue name (string)
                atomname: The atom name (string)
            Returns
                rname:    The forcefield's name for this residue (string)
                aname:    The forcefield's name for this atom (string)
        """
        rname = None
        aname = None
        if resname in self.map:
            res = self.map[resname]
            if res.hasAtom(atomname):
                atom = res.atoms[atomname]
                aname = atom.name
                rname = atom.resname
        return rname, aname

    def getGroup(self, resname, atomname):
        """
            Get the group/type associated with the input
            fields.  If not found, return a null string.
            
            Parameters:
                resname:  The residue name (string)
                atomname: The atom name (string)
        """
        group = ""
        if resname in self.map:
            resid = self.map[resname]
            if resid.hasAtom(atomname):
                atom = resid.atoms[atomname]
                group = atom.group
        return group

    def getParams(self, resname, atomname):
        """
            Get the parameters associated with the input fields.
            The residue itself is needed instead of simply its name
            because  the forcefield may use a different residue name
            than the standard amino acid name.

            Parameters
                resname:  The residue name (string)
                atomname: The atom name (string)
            Returns
                charge:   The charge on the atom (float)
                radius:   The radius of the atom (float)
        """
        charge = None
        radius = None

        if resname in self.map:
            resid = self.map[resname]
            if resid.hasAtom(atomname):
                atom = resid.atoms[atomname]
                charge = atom.charge
                radius = atom.radius
                
        return charge, radius

    def getParams1(self, residue, name):
        """
            Get the parameters associated with the input fields.
            The residue itself is needed instead of simply its name
            because  the forcefield may use a different residue name
            than the standard amino acid name.

            Parameters
                residue:  The residue (residue)
                name:     The atom name (string)
            Returns
                charge:   The charge on the atom (float)
                radius:   The radius of the atom (float)
        """
        charge = None
        radius = None
        resname = ""
        atomname = ""


        if self.name == "amber":
            resname, atomname = self.getAmberParams(residue, name)
        elif self.name == "charmm":
            resname, atomname = self.getCharmmParams(residue, name)
        elif self.name == "parse":
            resname, atomname = self.getParseParams(residue, name)
        else:
            resname = residue.name
            atomname = name

        defresidue = self.getResidue(resname)
#        print "resname: %s, defresidue: %s" % (resname, defresidue)
        if defresidue == None:
            return charge, radius

        atom = defresidue.getAtom(atomname)
        if atom != None:
            charge = atom.get("charge")
            radius = atom.get("radius")

        return charge, radius

    def getAmberParams(self, residue, name):
        """
            Get the forcefield definitions from the Amber database

            Parameters
                residue:  The residue (residue)
                name:     The atom name (string)
            Returns
                resname:  The name of the amber residue
                atomname: The name of the amber atom
        """
        atomname = name
        type = residue.get("type")
        if type == 4:
            resname = residue.get("naname")
        else:
            resname = residue.get("name")
        
        # Residue Substitutions
            
        if residue.get("name") == "CYS" and "HG" not in residue.get("map"):
            resname = "CYX"
        elif residue.get("name") == "HIS":
            if "HD1" in residue.get("map") and "HE2" in residue.get("map"):
                resname = "HIP"
            elif "HD1" in residue.get("map"):
                resname = "HID"
            elif "HE2" in residue.get("map"):
                resname = "HIE"
            else:
                resname = "HID" # Default for no hydrogens
        elif residue.get("name") == "HSP":
            resname = "HIP"
        elif residue.get("name") == "HSE":
            resname = "HIE"
        elif residue.get("name") == "HSD":
            resname = "HID"
        elif residue.get("name") == "GLU" or residue.get("name") == "GLH":
            if "HE1" in residue.get("map"):
                resname = "GLH"
                if atomname == "HE1": atomname = "HE2"
                elif atomname == "OE1": atomname = "OE2"
                elif atomname == "OE2": atomname = "OE1"
            elif "HE2" in residue.get("map"): resname = "GLH"
        elif residue.get("name") == "ASP" or residue.get("name") == "ASH":
            if "HD1" in residue.get("map"):
                resname = "ASH"
                if atomname == "HD1": atomname = "HD2"
                elif atomname == "OD1": atomname = "OD2"
                elif atomname == "OD2": atomname = "OD1"
            elif "HD2" in residue.get("map"): resname = "ASH"
               
        if residue.get("isCterm"):
            resname = "C" + resname
        elif residue.get("isNterm"):
            resname = "N" + resname

        # Atom Substitutions
  
        if resname == "WAT":
            if atomname == "O": atomname = "OW"
            elif atomname == "H1": atomname = "HW"
            elif atomname == "H2": atomname = "HW"
        elif resname == "ILE":
            if atomname == "CD": atomname = "CD1"
        if resname[0] == "N" and resname != "NME": # N-terminal
            if atomname == "H": atomname = "H1"
        if (resname == "CCYS" or resname == "NCYS") and atomname == "HG": atomname = "HSG"
        if resname == "CYM" and atomname == "H": atomname = "HN"
        if residue.get("isNterm") and resname == "NPRO" and atomname == "HN2":
            atomname = "H2"
        if residue.get("isNterm") and resname == "NPRO" and atomname == "HN1":
            atomname = "H3"   
        return resname, atomname

    def getParseParams(self, residue, name):
        """
            Get the forcefield definitions from the Parse database

            Parameters
                residue:  The residue (residue)
                name:     The atom name (string)
            Returns
                resname:  The name of the amber residue
                atomname: The name of the amber atom
        """
        atomname = name
        resname = residue.name

        # Terminal/Water Substitutions

        nterm = residue.get("isNterm")
        cterm = residue.get("isCterm")
        if nterm and resname != "ACE":
            if resname == "PRO" and nterm == 2:
                resname = "PR+"
                if atomname == "H2": atomname = "HN1"
                elif atomname == "H3": atomname = "HN2"
            elif resname == "PRO" and nterm == 1:
                resname = "PRN"
                if atomname == "H2" or atomname == "H3": atomname = "HN"
            elif nterm == 2: # Neutral
                if atomname in ["N","H","H2","H3","CA","HA","C","O"]:
                    resname = "BKN"
                if atomname == "H":
                    atomname = "H1"
                if atomname == 'H3':
                    atomname='H2'
            elif nterm == 3: # Positive
                if atomname in ["N","H","H2","H3","CA","HA","C","O"]:
                    resname = "BK+"
                if atomname == "H": atomname = "H1"
        elif cterm:
            if atomname == "O": atomname = "O1"
            elif atomname == "OXT": atomname = "O2"
            if cterm == 1 and atomname in ["N","H","HA","CA","C","O1","O2"]:
                resname = "BK-"
            elif cterm == 2 and atomname in ["N","H","HA","CA","C","O1","O2","HO"]:
                if atomname == "HO": atomname = "H2"
                resname = "BKC"
            #print 'Cterm resname is',resname
        elif residue.get("type") == 3:
            resname = "H2O"
            if atomname == "O": atomname = "OH"
            elif atomname == "H1": atomname = "HH1"
            elif atomname == "H2": atomname = "HH2"

        # Residue Substitutions
        if resname == "HSD": resname = "HID"
        elif resname in ["HIE","HSE"]: resname = "HIS"
        elif resname in ["HIP","HSP"]: resname = "HI+"
        elif resname == "ILE":
            if atomname == "HG12": atomname = "HG11"
            elif atomname == "HG13": atomname = "HG12"
            elif atomname == "CD": atomname = "CD1"
        elif resname == "CYS" and "HG" not in residue.get("map"):
            resname = "CSS"
        #
        # Histidine
        #
        elif resname == "HIS":
            if "HD1" in residue.get("map") and "HE2" in residue.get("map"):
                resname = "HI+"
            elif "HD1" in residue.get("map"):
                resname = "HID"
            elif "HE2" in residue.get("map"):
                resname = "HIS"
        elif resname == "GLU" or resname == "GLH":
            if "HE1" in residue.get("map"):
                resname = "GL0"
                if atomname == "HE1": atomname = "HE2"
                elif atomname == "OE1": atomname = "OE2"
                elif atomname == "OE2": atomname = "OE1"
            elif "HE2" in residue.get("map"): resname = "GL0"
        elif resname == "ASP" or resname == "ASH":
            if "HD1" in residue.get("map"):
                resname = "AS0"
                if atomname == "HD1": atomname = "HD2"
                elif atomname == "OD1": atomname = "OD2"
                elif atomname == "OD2": atomname = "OD1"
            elif "HD2" in residue.get("map"): resname = "AS0"
        elif resname == "ACE":
            if atomname == "HH31": atomname = "HA1"
            elif atomname == "HH32": atomname = "HA2"
            elif atomname == "HH33": atomname = "HA3"
            elif atomname == "CH3": atomname = "CA"    
        elif resname == "TYR":
            if not "HH" in residue.get("map"):
                resname="TYM"
        elif resname == "TYM": resname = "TY-"
        elif resname == "CYM": resname = "CY-"
        elif resname == "LYN": resname = "LY0"
        #
        # Neutral LYS and neutral ARG detection based on hydrogens - added by Jens
        #
        elif resname == "LYS":
            if not "HZ3" in residue.get("map"):
                resname="LY0"
        elif resname == "ARG":
            if not "HE" in residue.get("map"):
                resname="AR0"
        elif resname == "NME":
            resname = "N-M"
            if atomname == "CH3": atomname = "CA"
            elif atomname == "H": atomname = "H1"
            elif atomname.startswith("HH"): atomname = "HA" + atomname[-1]
        
        # Hydrogen Substitutions

        if atomname == "H": atomname = "HN"
        elif atomname == "HA2": atomname = "HA1"
        elif atomname == "HA3": atomname = "HA2"
        elif atomname == "HB2" and resname not in ["ALA"]: atomname = "HB1"
        elif atomname == "HB3" and resname not in ["ALA"]: atomname = "HB2"
        elif atomname == "HD2" and resname not in ["HIS","HI+","HID","AS0"]: atomname = "HD1"
        elif atomname == "HD3" and resname not in ["HIS","HI+","HID"]: atomname = "HD2"
        elif atomname == "HE2" and resname not in ["TRP","HIS","HI+","HID","GL0"]: atomname = "HE1"
        elif atomname == "HE3" and resname not in ["TRP","HIS","HI+","HID"]: atomname = "HE2"
        elif atomname == "HG2": atomname = "HG1"
        elif atomname == "HG3": atomname = "HG2"
        elif atomname == "HZ2" and resname == "LY0": atomname = "HZ1"
        elif atomname == "HZ3" and resname == "LY0": atomname = "HZ2"

        return resname, atomname

    def getCharmmParams(self, residue, name):
        """
            Get the forcefield definitions from the Charmm database

            Parameters
                residue:  The residue (residue)
                name:     The atom name (string)
            Returns
                resname:  The name of the Charmm residue
                atomname: The name of the Charmm atom
        """
        resname = residue.get("name")
        atomname = name

        #  Nucleic Acid Substitutions
        
        if residue.get("type") == 4:
            resname = resname[0]
            if resname == "A": resname = "ADE"
            elif resname == "C": resname = "CYT"
            elif resname == "G": resname = "GUA"
            elif resname == "T":
                resname = "THY"
                if atomname == "C7": atomname = "C5M"
                elif atomname == "H71": atomname = "H51"
                elif atomname == "H72": atomname = "H52"
                elif atomname == "H73": atomname = "H53" 
            elif resname == "U": resname = "URA"

            if atomname == "H5'1": atomname = "H5'"
            elif atomname == "H5'2": atomname = "H5''"
            elif atomname == "H2'1": atomname = "H2'"
            elif atomname in ["H2'2","HO'2"]: atomname = "H2''"
            
            if residue.getAtom("O2'") == None:
                if atomname in ["C2'","H2'","H2''"]: resname = "DEO1"

            if residue.getAtom("H5T") != None:
                if atomname in ["H5T","O5'","C5'"]: resname = "5TER"
            if residue.getAtom("H3T") != None:
                if atomname in ["H3T","O3'","C3'"]: resname = "3TER"
                
        # Terminal/Water Substitutions

        if residue.get("isNterm"):
            if resname == "GLY" and atomname in ["N","H","H2","H3","CA","HA2","HA3"]:
                resname = "GLYP"
                if atomname == "H": atomname = "HT1"
                elif atomname == "H2": atomname = "HT2"
                elif atomname == "H3": atomname = "HT3"
            elif resname == "PRO" and atomname in ["N","HN1","HN2","CD","CA","HD1","HD2","HA","H2","H3"]:
                resname = "PROP"
                if atomname == "H2": atomname = "HN1"
                elif atomname == "H3": atomname = "HN2"
            elif resname == "ACE":
                if atomname == "CH3": atomname = "CAY"
                elif atomname == "HH31": atomname = "HY1"
                elif atomname == "HH32": atomname = "HY2"
                elif atomname == "HH33": atomname = "HY3"
                elif atomname == "C": atomname = "CY"
                elif atomname == "O": atomname = "OY"
            else:
                if atomname in ["N","H","H2","H3","CA","HA"]:
                    resname = "NTER"
                    if atomname == "H": atomname = "HT1"
                    elif atomname == "H2": atomname = "HT2"
                    elif atomname == "H3": atomname = "HT3"               
        elif residue.get("isCterm"):
            if atomname in ["O","OXT","C"]:
                resname = "CTER"
                if atomname == "O":
                    atomname = "OT1"
                elif atomname == "OXT":
                    atomname = "OT2"
        elif residue.get("type") == 3:
            resname = "TP3M"
            if atomname == "O": atomname = "OH2"

        # Residue substitutions
            
        if resname == "ILE":
            if atomname == "CD1": atomname = "CD"
            elif atomname == "HD11": atomname = "HD1"
            elif atomname == "HD12": atomname = "HD2"
            elif atomname == "HD13": atomname = "HD3"
            elif atomname == "HG12": atomname = "HG11"
            elif atomname == "HG13": atomname = "HG12"
        elif resname == "CYS" and "HG" not in residue.get("map"):
            resname = "CYS"
            if atomname == "CB":
                resname = "DISU"
                atomname = "1CB"
            elif atomname == "SG":
                resname = "DISU"
                atomname = "1SG"
        elif resname == "HIS":
            if "HD1" in residue.get("map") and "HE2" in residue.get("map"):
                resname = "HSP"
            elif "HD1" in residue.get("map"):
                resname = "HSD"
            elif "HE2" in residue.get("map"):
                resname = "HSE"
        elif resname == "GLU" or resname == "GLH":
            if "HE1" in residue.get("map"):
                if atomname == "HE1": atomname = "HE2"
                elif atomname == "OE1": atomname = "OE2"
                elif atomname == "OE2": atomname = "OE1"
                if atomname in ["CG","HG3","HG1","HG2","CD","OE1","OE2","HE2"]: resname = "GLUP"
                else: resname = "GLU"
            elif "HE2" in residue.get("map"):
                if atomname in ["CG","HG3","HG1","HG2","CD","OE1","OE2","HE2"]: resname = "GLUP"
                else: resname = "GLU"
        elif resname == "ASP" or resname == "ASH":
            if "HD1" in residue.get("map"):
                if atomname == "HD1": atomname = "HD2"
                elif atomname == "OD1": atomname = "OD2"
                elif atomname == "OD2": atomname = "OD1"
                if atomname in ["CB","HB3","HB1","HB2","CG","OD1","OD2","HD2"]: resname = "ASPP"
                else: resname = "ASP"
            elif "HD2" in residue.get("map"):
                if atomname in ["CB","HB3","HB1","HB2","CG","OD1","OD2","HD2"]: resname = "ASPP"
                else: resname = "ASP"


        # HETATM Substitutions

        if resname == "ACE":
            if atomname == "CH3": atomname = "CAY"
            elif atomname == "HH31": atomname = "HY1"
            elif atomname == "HH32": atomname = "HY2"
            elif atomname == "HH33": atomname = "HY3"
            elif atomname == "C": atomname = "CY"
            elif atomname == "O": atomname = "OY"
        elif resname == "ADP":
            atomname = string.replace(atomname,"*","\'")
        elif resname == "NME":
            resname = "CT3"
            if atomname == "HH31": atomname = "HT1"
            elif atomname == "HH32": atomname = "HT2"
            elif atomname == "HH33": atomname = "HT3"
            elif atomname == "CH3": atomname = "CAT"
            elif atomname == "N": atomname = "NT"
            elif atomname == "H": atomname = "HNT"
            
        # Hydrogen Substitutions

        if atomname == "H": atomname = "HN"
        elif atomname == "HA2": atomname = "HA1"
        elif atomname == "HA3": atomname = "HA2"
        elif atomname == "HB2" and resname not in ["ALA"]: atomname = "HB1"
        elif atomname == "HB3" and resname not in ["ALA"]: atomname = "HB2"
        elif atomname == "HD2" and resname not in ["HSP","HSE","HSD","ASPP"]: atomname = "HD1"
        elif atomname == "HD3" and resname not in ["HIS","HSE","HSD"]: atomname = "HD2"
        elif atomname == "HE2" and resname not in ["TRP","HSP","HSE","HSD","GLUP"]: atomname = "HE1"
        elif atomname == "HE3" and resname not in ["TRP","HSP","HSE","HSD"]: atomname = "HE2"
        elif atomname == "HG2": atomname = "HG1"
        elif atomname == "HG3": atomname = "HG2"
        elif atomname == "HG" and resname in ["SER","CYS"]: atomname = "HG1"

        return resname, atomname

class ForcefieldResidue:
    """
        ForcefieldResidue class

        The ForceFieldResidue class contains a mapping of all atoms within
        the residue for easy searching.
    """
    def __init__(self, name):
        """
            Initialize the ForceFieldResidue object

            Parameters
                name: The name of the residue (string)
        """
        self.name = name
        self.atoms = {}


    def addAtom(self, atom):
        """
            Add an atom to the ForcefieldResidue

            Parameters
                atom:  The atom to be added (atom)
        """
        atomname = atom.get("name")
        self.atoms[atomname] = atom

    def getAtoms(self):
        """
            Return the list of atoms in this residue.
        """
        return self.atoms

    def hasAtom(self, atomname):
        """
            Check to see if the atomname is in the current residue.

            Parameters
                atomname:  The name of the atom to search for
            Returns
                1 if the atom is present in the residue, 0 otherwise
        """
        if atomname in self.atoms: return 1
        else: return 0

    def getAtom(self, atomname):
        """
            Return the atom object with the given atomname

            Parameters
                resname: The name of the atom (string)
            Returns
                residue: The atom object (ForcefieldAtom)
        """
        if self.hasAtom(atomname): return self.atoms[atomname]
        else: return None

class ForcefieldAtom:
    """
        ForcefieldAtom class

        The ForcefieldAtom object contains fields that are related to the
        forcefield at the atom level
    """
    
    def __init__(self, name, charge, radius, resname, group=""):
        """
            Initialize the object

            Parameters
                name:    The atom name (string)
                charge:  The charge on the atom (float)
                radius:  The radius of the atom (float)
                resname: The residue name (string)
                group:   The group name (string)
        """
        self.name = name
        self.charge = charge
        self.radius = radius
        self.resname = resname
        self.group = group

    def get(self, name):
        """
            Get a member of the ForcefieldAtom class

            Parameters
                name:       The name of the member (string)
            Possible Values
                name:    The atom name (string)
                charge:  The charge on the atom (float)
                radius:  The radius of the atom (float)
                epsilon: The epsilon assocaited with the atom (float)
            Returns
                item:       The value of the member
        """
        try:
            item = getattr(self, name)
            return item
        except AttributeError:
            message = "Unable to access object \"%s\" in class ForcefieldAtom" % name
            raise PDBInternalError(message)

    def __str__(self):
        """
            String representation of the forcefield atom.
        """
        txt = "%s:\n"% self.name
        txt += "  Charge: %.4f\n" % self.charge
        txt += "  Radius: %.4f" % self.radius
        return txt



import sys

sys.path.append("/Users/nielsen/bin/APBS/tools/python/vgrid")
from vgrid import *

NMAX = 5


class phimap_class:
    def __init__(self, xmap, ymap, zmap):
        # print
        # print 'Reading initial Epsmaps'
        # print
        self.maps = {}
        for name, emap in [["x", xmap], ["y", ymap], ["z", zmap]]:
            print "Reading", emap
            data = []
            value = 0.0

            startVio()
            import sys
            from sys import stdout, stderr

            grid = Vgrid_ctor(0, 0, 0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, data)
            Vgrid_readDX(grid, "FILE", "ASC", "", emap)

            nx = grid.nx
            ny = grid.ny
            nz = grid.nz
            hx = grid.hx
            hy = grid.hy
            hzed = grid.hzed
            xmin = grid.xmin
            ymin = grid.ymin
            zmin = grid.zmin
            # Get the grid data
            grid_data = []
            for k in range(nz):
                for j in range(ny):
                    for i in range(nx):
                        inval = 0.0
                        pt = [xmin + i * hx, ymin + j * hy, zmin + k * hzed]
                        ret, value = Vgrid_value(grid, pt, inval)
                        grid_data.append(value)
                        if not ret:
                            raise Exception("Grid point not found")
        return grid_data

    def writemap(self, map, filename="tester.dx"):
        """Write this map in dx format"""
        mydata = []
        # count=0
        for k in range(map.nz):
            for j in range(map.ny):
                for i in range(map.nx):
                    mydata.append(map.get_value(i, j, k))
        #
        # Create the grid
        #
        # print mydata
        mygrid = Vgrid_ctor(
            map.nx,
            map.ny,
            map.nz,
            map.hx,
            map.hy,
            map.hz,
            map.xmin,
            map.ymin,
            map.zmin,
            mydata,
        )
        Vgrid_writeDX(mygrid, "FILE", "ASC", "", filename, "What is this", null_array())
        delete_vgrid(mygrid)
        return



"""
    Driver for PDB2PQR

    This module takes a PDB file as input and performs optimizations
    before yielding a new PDB-style file as output.

    Ported to Python by Todd Dolinsky (todd@ccb.wustl.edu)
    Washington University in St. Louis

    Parsing utilities provided by Nathan A. Baker (Nathan.Baker@pnl.gov)
    Pacific Northwest National Laboratory

    Copyright (c) 2002-2011, Jens Erik Nielsen, University College Dublin;
    Nathan A. Baker, Battelle Memorial Institute, Developed at the Pacific
    Northwest National Laboratory, operated by Battelle Memorial Institute,
    Pacific Northwest Division for the U.S. Department Energy.;
    Paul Czodrowski & Gerhard Klebe, University of Marburg.

	All rights reserved.

	Redistribution and use in source and binary forms, with or without modification,
	are permitted provided that the following conditions are met:

		* Redistributions of source code must retain the above copyright notice,
		  this list of conditions and the following disclaimer.
		* Redistributions in binary form must reproduce the above copyright notice,
		  this list of conditions and the following disclaimer in the documentation
		  and/or other materials provided with the distribution.
        * Neither the names of University College Dublin, Battelle Memorial Institute,
          Pacific Northwest National Laboratory, US Department of Energy, or University
          of Marburg nor the names of its contributors may be used to endorse or promote
          products derived from this software without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
	ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
	IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
	INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
	BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
	DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
	LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
	OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
	OF THE POSSIBILITY OF SUCH DAMAGE.

"""

__date__  = "5 April 2010"
__author__ = "Todd Dolinsky, Nathan Baker, Jens Nielsen, Paul Czodrowski, Jan Jensen, Samir Unni, Yong Huang"

import string
import sys
import getopt
from optparse import OptionParser, OptionGroup
import os
import time
import copy
from src import pdb
from src import utilities
from src import structures
from src import routines
from src import protein
from src.pdb import *
from src.utilities import *
from src.structures import *
from src.definitions import *
from src.forcefield import *
from src.routines import *
from src.protein import *
from src.server import *
from src.hydrogens import *
from src.aconf import *
from StringIO import *
from src.errors import PDB2PQRError

__version__ = PDB2PQR_VERSION

import extensions

def getOldHeader(pdblist):
    oldHeader = StringIO()
    headerTypes = (HEADER, TITLE, COMPND, SOURCE,
                   KEYWDS, EXPDTA, AUTHOR, REVDAT,
                   JRNL, REMARK, SPRSDE, NUMMDL)
    for pdbObj in pdblist:
        if not isinstance(pdbObj,headerTypes):
            break

        oldHeader.write(str(pdbObj))
        oldHeader.write('\n')

    return oldHeader.getvalue()

def printPQRHeader(pdblist,
                   atomlist,
                   reslist,
                   charge,
                   ff,
                   warnings,
                   ph_calc_method,
                   pH,
                   ffout,
                   cl_args,
                   include_old_header = False):
    """
        Print the header for the PQR file

        Parameters:
            atomlist: A list of atoms that were unable to have
                      charges assigned (list)
            reslist:  A list of residues with non-integral charges
                      (list)
            charge:   The total charge on the protein (float)
            ff:       The forcefield name (string)
            warnings: A list of warnings generated from routines (list)
            pH :  pH value, if any. (float)
            ffout :  ff used for naming scheme (string)
            cl_args : the command line argument used when running pdb2pqr (string)
        Returns
            header:   The header for the PQR file (string)
    """
    if ff is None:
        ff = 'User force field'
    else:
        ff = ff.upper()
    header = "REMARK   1 PQR file generated by PDB2PQR (Version %s)\n" % __version__
    header = header + "REMARK   1\n"
    if cl_args is not None:
        header +=     "REMARK   1 Command line used to generate this file:\n"
        header +=     "REMARK   1 " + cl_args + "\n"
        header = header + "REMARK   1\n"
    header = header + "REMARK   1 Forcefield Used: %s\n" % ff
    if not ffout is None:
        header = header + "REMARK   1 Naming Scheme Used: %s\n" % ffout
    header = header + "REMARK   1\n"

    if ph_calc_method is not None:
        header = header + "REMARK   1 pKas calculated by %s and assigned using pH %.2f\n" % (ph_calc_method, pH)
        header = header + "REMARK   1\n"

    for warning in warnings:
        header = header + "REMARK   5 " + warning
    header = header + "REMARK   5\n"

    if len(atomlist) != 0:
        header += "REMARK   5 WARNING: PDB2PQR was unable to assign charges\n"
        header += "REMARK   5          to the following atoms (omitted below):\n"
        for atom in atomlist:
            header += "REMARK   5              %i %s in %s %i\n" % \
                      (atom.get("serial"), atom.get("name"), \
                       atom.get("residue").get("name"), \
                       atom.get("residue").get("resSeq"))
        header += "REMARK   5 This is usually due to the fact that this residue is not\n"
        header += "REMARK   5 an amino acid or nucleic acid; or, there are no parameters\n"
        header += "REMARK   5 available for the specific protonation state of this\n"
        header += "REMARK   5 residue in the selected forcefield.\n"
        header += "REMARK   5\n"
    if len(reslist) != 0:
        header += "REMARK   5 WARNING: Non-integral net charges were found in\n"
        header += "REMARK   5          the following residues:\n"
        for residue in reslist:
            header += "REMARK   5              %s - Residue Charge: %.4f\n" % \
                      (residue, residue.getCharge())
        header += "REMARK   5\n"
    header += "REMARK   6 Total charge on this protein: %.4f e\n" % charge
    header += "REMARK   6\n"

    if include_old_header:
        header += "REMARK   7 Original PDB header follows\n"
        header += "REMARK   7\n"

        header += getOldHeader(pdblist)

    return header

def runPDB2PQR(pdblist, ff,
               outname = "",
               ph = None,
               verbose = False,
               selectedExtensions = [],
               extensionOptions = utilities.ExtraOptions(),
               ph_calc_method = None,
               ph_calc_options = None,
               clean = False,
               neutraln = False,
               neutralc = False,
               ligand = None,
               assign_only = False,
               chain = False,
			   drop_water = False,
               debump = True,
               opt = True,
               typemap = False,
               userff = None,
               usernames = None,
               ffout = None,
               commandLine=None,
               include_old_header=False):
    """
        Run the PDB2PQR Suite

        Arguments:
            pdblist: The list of objects that was read from the PDB file
                     given as input (list)
            ff:      The name of the forcefield (string)

        Keyword Arguments:
            outname:       The name of the desired output file
            ph:            The desired ph of the system (float)
            verbose:       When True, script will print information to stdout
                             When False, no detailed information will be printed (float)
            extensions:      List of extensions to run
            extensionOptions:optionParser like option object that is passed to each object.
            propkaOptions:optionParser like option object for propka30.
            clean:         only return original PDB file in aligned format.
            neutraln:      Make the N-terminus of this protein neutral
            neutralc:      Make the C-terminus of this protein neutral
            ligand:        Calculate the parameters for the ligand in mol2 format at the given path.
            assign_only:   Only assign charges and radii - do not add atoms, debump, or optimize.
            chain:         Keep the chain ID in the output PQR file
            drop_water:    Remove water molecules from output
            debump:        When 1, debump heavy atoms (int)
            opt:           When 1, run hydrogen optimization (int)
            typemap:       Create Typemap output.
            userff:        The user created forcefield file to use. Overrides ff.
            usernames:     The user created names file to use. Required if using userff.
            ffout:         Instead of using the standard canonical naming scheme for residue and atom names,  +
                           use the names from the given forcefield
            commandLine:   command line used (if any) to launch the program. Included in output header.
            include_old_header: Include most of the PDB header in output.
            pdb2pka_params: parameters for running pdb2pka.

        Returns
            header:  The PQR file header (string)
            lines:   The PQR file atoms (list)
            missedligandresidues:  A list of ligand residue names whose charges could
                     not be assigned (ligand)
    """

    pkaname = ""
    lines = []
    Lig = None
    atomcount = 0   # Count the number of ATOM records in pdb

    outroot = utilities.getPQRBaseFileName(outname)

    if ph_calc_method == 'propka':
        pkaname = outroot + ".propka"
        #TODO: What? Shouldn't it be up to propka on how to handle this?
        if os.path.isfile(pkaname):
            os.remove(pkaname)

    start = time.time()

    if verbose:
        print "Beginning PDB2PQR...\n"

    myDefinition = Definition()
    if verbose:
        print "Parsed Amino Acid definition file."

    if drop_water:
        # Remove the waters
        pdblist_new = []
        for record in pdblist:
            if isinstance(record, (HETATM, ATOM, SIGATM, SEQADV)):
                if record.resName in WAT.water_residue_names:
                    continue
            pdblist_new.append(record)

        pdblist = pdblist_new

    # Check for the presence of a ligand!  This code is taken from pdb2pka/pka.py

    if not ligand is None:
        from pdb2pka.ligandclean import ligff
        myProtein, myDefinition, Lig = ligff.initialize(myDefinition, ligand, pdblist, verbose)
        for atom in myProtein.getAtoms():
            if atom.type == "ATOM":
                atomcount += 1
    else:
        myProtein = Protein(pdblist, myDefinition)

    if verbose:
        print "Created protein object -"
        print "\tNumber of residues in protein: %s" % myProtein.numResidues()
        print "\tNumber of atoms in protein   : %s" % myProtein.numAtoms()

    myRoutines = Routines(myProtein, verbose)

    for residue in myProtein.getResidues():
        multoccupancy = 0
        for atom in residue.getAtoms():
            if atom.altLoc != "":
                multoccupancy = 1
                txt = "Warning: multiple occupancies found: %s in %s\n" % (atom.name, residue)
                sys.stderr.write(txt)
        if multoccupancy == 1:
            myRoutines.warnings.append("WARNING: multiple occupancies found in %s,\n" % (residue))
            myRoutines.warnings.append("         at least one of the instances is being ignored.\n")

    myRoutines.setTermini(neutraln, neutralc)
    myRoutines.updateBonds()

    if clean:
        header = ""
        lines = myProtein.printAtoms(myProtein.getAtoms(), chain)

        # Process the extensions
        for ext in selectedExtensions:
            module = extensions.extDict[ext]
            #TODO: figure out a way to do this without crashing...
            #tempRoutines = copy.deepcopy(myRoutines)
            module.run_extension(myRoutines, outroot, extensionOptions)

        if verbose:
            print "Total time taken: %.2f seconds\n" % (time.time() - start)

        #Be sure to include None for missed ligand residues
        return header, lines, None

    #remove any future need to convert to lower case
    if not ff is None:
        ff = ff.lower()
    if not ffout is None:
        ffout = ffout.lower()

    if not assign_only:
        # It is OK to process ligands with no ATOM records in the pdb
        if atomcount == 0 and Lig != None:
            pass
        else:
            myRoutines.findMissingHeavy()
        myRoutines.updateSSbridges()

        if debump:
            myRoutines.debumpProtein()

        if ph_calc_method == 'propka':
            myRoutines.runPROPKA(ph, ff, outroot, pkaname, ph_calc_options)
        elif ph_calc_method == 'pdb2pka':
            myRoutines.runPDB2PKA(ph, ff, pdblist, ligand, verbose, ph_calc_options)

        myRoutines.addHydrogens()

        myhydRoutines = hydrogenRoutines(myRoutines)

        if debump:
            myRoutines.debumpProtein()

        if opt:
            myhydRoutines.setOptimizeableHydrogens()
            myhydRoutines.initializeFullOptimization()
            myhydRoutines.optimizeHydrogens()
        else:
            myhydRoutines.initializeWaterOptimization()
            myhydRoutines.optimizeHydrogens()

        # Special for GLH/ASH, since both conformations were added
        myhydRoutines.cleanup()

    else:  # Special case for HIS if using assign-only
        for residue in myProtein.getResidues():
            if isinstance(residue, HIS):
                myRoutines.applyPatch("HIP", residue)

    myRoutines.setStates()

    myForcefield = Forcefield(ff, myDefinition, userff, usernames)
    hitlist, misslist = myRoutines.applyForcefield(myForcefield)
    

    ligsuccess = 0

    if not ligand is None:
        # If this is independent, we can assign charges and radii here
        for residue in myProtein.getResidues():
            if isinstance(residue, LIG):
                templist = []
                Lig.make_up2date(residue)
                for atom in residue.getAtoms():
                    atom.ffcharge = Lig.ligand_props[atom.name]["charge"]
                    atom.radius = Lig.ligand_props[atom.name]["radius"]
                    if atom in misslist:
                        misslist.pop(misslist.index(atom))
                        templist.append(atom)

                charge = residue.getCharge()
                if abs(charge - int(charge)) > 0.001:
                    # Ligand parameterization failed
                    myRoutines.warnings.append("WARNING: PDB2PQR could not successfully parameterize\n")
                    myRoutines.warnings.append("         the desired ligand; it has been left out of\n")
                    myRoutines.warnings.append("         the PQR file.\n")
                    myRoutines.warnings.append("\n")

                    # remove the ligand
                    myProtein.residues.remove(residue)
                    for myChain in myProtein.chains:
                        if residue in myChain.residues: myChain.residues.remove(residue)
                else:
                    ligsuccess = 1
                    # Mark these atoms as hits
                    hitlist = hitlist + templist

    # Temporary fix; if ligand was successful, pull all ligands from misslist
    if ligsuccess:
        templist = misslist[:]
        for atom in templist:
            if isinstance(atom.residue, (Amino, Nucleic)):
                continue
            misslist.remove(atom)

    # Create the Typemap
    if typemap:
        typemapname = "%s-typemap.html" % outroot
        myProtein.createHTMLTypeMap(myDefinition, typemapname)

    # Grab the protein charge
    reslist, charge = myProtein.getCharge()

    # If we want a different naming scheme, use that

    if not ffout is None:
        scheme = ffout
        userff = None # Currently not supported
        if scheme != ff:
            myNameScheme = Forcefield(scheme, myDefinition, userff)
        else:
            myNameScheme = myForcefield
        myRoutines.applyNameScheme(myNameScheme)

    header = printPQRHeader(pdblist, misslist, reslist, charge, ff,
                            myRoutines.getWarnings(), ph_calc_method, ph, ffout, commandLine,
                            include_old_header=include_old_header)
    lines = myProtein.printAtoms(hitlist, chain)

    # Determine if any of the atoms in misslist were ligands
    missedligandresidues = []
    for atom in misslist:
        if isinstance(atom.residue, (Amino, Nucleic)):
            continue
        if atom.resName not in missedligandresidues:
            missedligandresidues.append(atom.resName)

    # Process the extensions
    for ext in selectedExtensions:
        module = extensions.extDict[ext]
        #TODO: figure out a way to do this without crashing...
        #tempRoutines = copy.deepcopy(myRoutines)
        module.run_extension(myRoutines, outroot, extensionOptions)


    if verbose:
        print "Total time taken: %.2f seconds\n" % (time.time() - start)

    return header, lines, missedligandresidues

def mainCommand(argv):
    """
        Main driver for running program from the command line.
    """

    fieldNames = ('amber','charmm','parse', 'tyl06','peoepb','swanson', 'gromos')

    validForcefields = []
    validForcefields.extend(fieldNames)
    validForcefields.extend((x.upper() for x in fieldNames))

    description = 'This module takes a PDB file as input and performs ' +\
                  'optimizations before yielding a new PQR-style file in PQR_OUTPUT_PATH.\n' +\
                  'If PDB_PATH is an ID it will automatically be obtained from the PDB archive.'

    usage = 'Usage: %prog [options] PDB_PATH PQR_OUTPUT_PATH'

    parser = OptionParser(description=description, usage=usage, version='%prog (Version ' + __version__ + ')')


    group = OptionGroup(parser,"Manditory options", "One of the following options must be used.")
    group.add_option('--ff', dest='ff', metavar='FIELD_NAME', choices=validForcefields,
                      help='The forcefield to use - currently amber, ' +
                           'charmm, parse, tyl06, peoepb and swanson ' +
                           'are supported.')

    group.add_option('--userff', dest='userff', metavar='USER_FIELD_FILE',
                      help='The user created forcefield file to use. Requires --usernames overrides --ff')

    group.add_option('--clean', dest='clean', action='store_true', default=False,
                      help='Do no optimization, atom addition, or parameter assignment, ' +
                           'just return the original PDB file in aligned format. ' +
                           'Overrides --ff and --userff')
    parser.add_option_group(group)


    group = OptionGroup(parser,"General options")
    group.add_option('--nodebump', dest='debump', action='store_false', default=True,
                      help='Do not perform the debumping operation')

    group.add_option('--noopt', dest='opt', action='store_false', default=True,
                      help='Do not perform hydrogen optimization')

    group.add_option('--chain', dest='chain', action='store_true', default=False,
                      help='Keep the chain ID in the output PQR file')

    group.add_option('--assign-only', dest='assign_only', action='store_true', default=False,
                      help='Only assign charges and radii - do not add atoms, debump, or optimize.')

    group.add_option('--ffout', dest='ffout', metavar='FIELD_NAME',choices=validForcefields,
                      help='Instead of using the standard canonical naming scheme for residue and atom names, ' +
                           'use the names from the given forcefield - currently amber, ' +
                           'charmm, parse, tyl06, peoepb and swanson are supported.')

    group.add_option('--usernames', dest='usernames', metavar='USER_NAME_FILE',
                      help='The user created names file to use. Required if using --userff')

    group.add_option('--apbs-input', dest='input', action='store_true', default=False,
                      help='Create a template APBS input file based on the generated PQR file.  Also creates a Python ' +
                           'pickle for using these parameters in other programs.')

    group.add_option('--ligand', dest='ligand',  metavar='PATH',
                      help='Calculate the parameters for the ligand in mol2 format at the given path. ' +
                           'Pdb2pka must be compiled.')

    group.add_option('--whitespace', dest='whitespace', action='store_true', default=False,
                      help='Insert whitespaces between atom name and residue name, between x and y, and between y and z.')

    group.add_option('--typemap', dest='typemap', action='store_true', default=False,
                      help='Create Typemap output.')

    group.add_option('--neutraln', dest='neutraln', action='store_true', default=False,
                      help='Make the N-terminus of this protein neutral (default is charged). '
                           'Requires PARSE force field.')

    group.add_option('--neutralc', dest='neutralc', action='store_true', default=False,
                      help='Make the C-terminus of this protein neutral (default is charged). '
                           'Requires PARSE force field.')

    group.add_option('-v', '--verbose', dest='verbose', action='store_true', default=False,
                      help='Print information to stdout.')

    group.add_option('--drop-water', dest='drop_water', action='store_true', default=False,
                      help='Drop waters before processing protein. Currently recognized and deleted are the following water types:  %s' % ', '.join(WAT.water_residue_names))

    group.add_option('--include-header', dest='include_header', action='store_true', default=False,
                      help='Include pdb header in pqr file. '
                           'WARNING: The resulting PQR file will not work with APBS versions prior to 1.5')
    parser.add_option_group(group)

    pka_group = OptionGroup(parser,"pH options")

    pka_group.add_option('--ph-calc-method', dest='ph_calc_method', metavar='PH_METHOD', choices=('propka', 'pdb2pka'),
                      help='Method used to calculate ph values. If a pH calculation method is selected, for each'
                      ' titratable residue pH values will be calculated and the residue potentially modified'
                      ' after comparison with the pH value supplied by --with_ph\n'
                      'propka - Use PROPKA to calculate pH values. Actual PROPKA results will be output to <output-path>.propka.\n'
                      'pdb2pka - Use PDB2PKA to calculate pH values. Requires the use of the PARSE force field.'
                      ' Warning: Larger residues can take a very long time to run using this method. EXPERIMENTAL!')

    pka_group.add_option('--with-ph', dest='ph', action='store', type='float', default=7.0,
                      help='pH values to use when applying the results of the selected pH calculation method.'
                      ' Defaults to %default')

    parser.add_option_group(pka_group)

    pdb2pka_group = OptionGroup(parser,"PDB2PKA method options")

    pdb2pka_group.add_option('--pdb2pka-out', dest='pdb2pka_out', action='store', default='pdb2pka_output',
                             help='Output directory for PDB2PKA results. Defaults to %default')
    pdb2pka_group.add_option('--pdb2pka-resume', dest='pdb2pka_resume', action="store_true", default=False,
                             help='Resume run from state saved in output directory.')

    pdb2pka_group.add_option('--pdie', dest='pdb2pka_pdie', default=8,type='int',
                             help='Protein dielectric constant. Defaults to %default')
    pdb2pka_group.add_option('--sdie', dest='pdb2pka_sdie', default=80, type='int',
                             help='Solvent dielectric constant. Defaults to %default')

#     pdb2pka_group.add_option('--maps', dest='maps', default=None, type='int',
#                              help='<1 for using provided 3D maps; 2 for genereting new maps>')
#     pdb2pka_group.add_option('--xdiel', dest='xdiel', default=None, type='str',
#                              help='<xdiel maps>')
#     pdb2pka_group.add_option('--ydiel', dest='ydiel', default=None, type='str',
#                              help='<ydiel maps>')
#     pdb2pka_group.add_option('--zdiel', dest='zdiel', default=None, type='str',
#                              help='<zdiel maps>')
#     pdb2pka_group.add_option('--kappa', dest='kappa', default=None, type='str',
#                              help='<ion-accessibility map>')
#     pdb2pka_group.add_option('--smooth', dest='sd', default=None, type='float',
#                              help='<st.dev [A] of Gaussian smooting of 3D maps at the boundary, bandthwith=3 st.dev>')
    #
    # Cut off energy for calculating non-charged-charged interaction energies
    #
    pdb2pka_group.add_option('--pairene',dest='pdb2pka_pairene',type='float',default=1.0,
                      help='Cutoff energy in kT for calculating non charged-charged interaction energies. Default: %default')

    parser.add_option_group(pdb2pka_group)

    propka_group = OptionGroup(parser,"PROPKA method options")

    propka_group.add_option("--propka-reference", dest="propka_reference", default="neutral", choices=('neutral','low-pH'),
           help="Setting which reference to use for stability calculations. See PROPKA 3.0 documentation.")

    propka_group.add_option('--propka-verbose', dest='propka_verbose', action='store_true', default=False,
                      help='Print extra proPKA information to stdout. '
                           'WARNING: This produces an incredible amount of output.')

    parser.add_option_group(propka_group)


    extensions.setupExtensionsOptions(parser)

    (options, args) = parser.parse_args(argv[1:])

    commandLine = ' '.join(argv[1:])

    if len(args) != 2:
        parser.error('Incorrect number (%d) of arguments!\nargs: %s' % (len(args), args))

    if options.assign_only or options.clean:
        options.debump = options.optflag = False

    userfffile = None
    usernamesfile = None

    if not options.clean:
        if not options.usernames is None:
            try:
                usernamesfile = open(options.usernames, 'rU')
            except IOError:
                parser.error('Unable to open user names file %s' % options.usernames)

        if not options.userff is None:
            try:
                userfffile = open(options.userff, 'rU')
            except IOError:
                parser.error('Unable to open user force field file %s' % options.userff)

            if options.usernames is None:
                parser.error('--usernames must be specified if using --userff')

        else:
            if options.ff is None:
                parser.error('One of the manditory options was not specified.\n' +
                             'Please specify either --ff, --userff, or --clean')

            if getFFfile(options.ff) == '':
                parser.error('Unable to find parameter files for forcefield %s!' % options.ff)

    if options.ph < 0.0 or options.ph > 14.0:
        parser.error('%i is not a valid pH!  Please choose a pH between 0.0 and 14.0.' % options.pH)

    ph_calc_options = None
    if options.ph_calc_method == 'propka':
        ph_calc_options = utilities.createPropkaOptions(options.ph,
                                                   verbose=options.propka_verbose,
                                                   reference=options.propka_reference)


    if options.ph_calc_method == 'pdb2pka':
        if options.ff.lower() != 'parse':
            parser.error('PDB2PKA requires the PARSE force field.')
        ph_calc_options = {'output_dir': options.pdb2pka_out,
                          'clean_output': not options.pdb2pka_resume,
                          'pdie': options.pdb2pka_pdie,
                          'sdie': options.pdb2pka_sdie,
                          'pairene': options.pdb2pka_pairene}

    if options.ligand is not None:
        try:
            options.ligand = open(options.ligand, 'rU')
        except IOError:
            parser.error('Unable to find ligand file %s!' % options.ligand)

    if options.neutraln and (options.ff is None or options.ff.lower() != 'parse'):
        parser.error('--neutraln option only works with PARSE forcefield!')

    if options.neutralc and (options.ff is None or options.ff.lower() != 'parse'):
        parser.error('--neutralc option only works with PARSE forcefield!')

    text =  """
--------------------------
PDB2PQR - a Python-based structural conversion utility
--------------------------
Please cite your use of PDB2PQR as:
  Dolinsky TJ, Nielsen JE, McCammon JA, Baker NA.
  PDB2PQR: an automated pipeline for the setup, execution,
  and analysis of Poisson-Boltzmann electrostatics calculations.
  Nucleic Acids Research 32 W665-W667 (2004).

"""
    sys.stdout.write(text)

    path = args[0]
    pdbFile = getPDBFile(path)
    pdblist, errlist = readPDB(pdbFile)

    if len(pdblist) == 0 and len(errlist) == 0:
        parser.error("Unable to find file %s!" % path)

    if len(errlist) != 0 and options.verbose:
        print "Warning: %s is a non-standard PDB file.\n" % path
        print errlist

    outpath = args[1]
    options.outname = outpath

    #In case no extensions were specified or no extensions exist.
    if not hasattr(options, 'active_extensions' ) or options.active_extensions is None:
        options.active_extensions = []

    #I see no point in hiding options from extensions.
    extensionOpts = options

    #TODO: The ideal would be to pass a file like object for the second
    # argument and add a third for names then
    # get rid of the userff and username arguments to this function.
    # This would also do away with the redundent checks and such in
    # the Forcefield constructor.
    try:
        header, lines, missedligands = runPDB2PQR(pdblist,
                                                  options.ff,
                                                  outname = options.outname,
                                                  ph = options.ph,
                                                  verbose = options.verbose,
                                                  selectedExtensions = options.active_extensions,
                                                  ph_calc_method = options.ph_calc_method,
                                                  ph_calc_options = ph_calc_options,
                                                  extensionOptions = extensionOpts,
                                                  clean = options.clean,
                                                  neutraln = options.neutraln,
                                                  neutralc = options.neutralc,
                                                  ligand = options.ligand,
                                                  assign_only = options.assign_only,
                                                  chain = options.chain,
                                                  drop_water = options.drop_water,
                                                  debump = options.debump,
                                                  opt = options.opt,
                                                  typemap = options.typemap,
                                                  userff = userfffile,
                                                  usernames = usernamesfile,
                                                  ffout = options.ffout,
                                                  commandLine = commandLine,
                                                  include_old_header = options.include_header)
    except PDB2PQRError as er:
        print er
        sys.exit(2)

    # Print the PQR file
    outfile = open(outpath,"w")
    outfile.write(header)
    # Adding whitespaces if --whitespace is in the options
    for line in lines:
        if options.whitespace:
            if line[0:4] == 'ATOM':
                newline = line[0:6] + ' ' + line[6:16] + ' ' + line[16:38] + ' ' + line[38:46] + ' ' + line[46:]
                outfile.write(newline)
            elif line[0:6] == 'HETATM':
                newline = line[0:6] + ' ' + line[6:16] + ' ' + line[16:38] + ' ' + line[38:46] + ' ' + line[46:]
                outfile.write(newline)
        else:
            outfile.write(line)
    outfile.close()

    if options.input:
        from src import inputgen
        from src import psize
        method = "mg-auto"
        size = psize.Psize()
        size.parseInput(outpath)
        size.runPsize(outpath)
        async = 0 # No async files here!
        input = inputgen.Input(outpath, size, method, async, potdx=True)
        input.printInputFiles()
        input.dumpPickle()


if __name__ == "__main__":
    mainCommand(sys.argv)



"""
    Summary extension

    Print protein summary file. 
    Currently prints a list of all residue in protein.
"""

__date__ = "21 October 2011"
__author__ = "Kyle Monson"

import extensions


def usage():
    """
    Returns usage text for summary.
    """
    return "Print protein summary information to {output-path}.summary."


def create_summary_output(routines, outfile):
    """
    Output the interaction energy between each possible residue pair.
    """
    routines.write("Printing protein summary...\n")

    output = extensions.extOutputHelper(routines, outfile)

    output.write(routines.protein.getSummary() + "\n")

    for residue in routines.protein.getResidues():
        output.write(str(residue) + "\n")


def run_extension(routines, outroot, options):
    outname = outroot + ".summary"
    with open(outname, "w") as outfile:
        create_summary_output(routines, outfile)
